//civ-compatable map
//resources (and update drawing)
//rivers and lakes

// I N C L U D E S ////////////////////////////////////////////////////////////

#include "./includes.h"
#include "./terrain.h"

// F U N C T I O N S //////////////////////////////////////////////////////////

void _2dfree(void **var, int width)
  {
  int a;
  for (a=0; a<width; a++)
    {
    free(var[a]);
    }
  free(var);
  }

float f_random()
  {
  return (((float)random())/RAND_MAX-0.5)*2.0;
  }

float uf_random()
  {
  return ((float)random())/RAND_MAX;
  }

float sf_random()	//Scaled float random (prefers numbers away from zero)
  {
  return sqrt(((float)random())/RAND_MAX)*((random()&1)*2-1);
  }

float suf_random()
  {
  return sqrt(((float)random())/RAND_MAX)*((random()&1)-0.5)+0.5;
  }

float square(float f)
  {
  return f*f;
  }

float de_mod(int a, int e, float c1, float c2)
  {
  float a2=(float)a;
  float e2=(float)e;
  float s=a2/e2;
  if (s<c1)
    {
    return 1.0;
    }
  else if (s<c1+c2)
    {
    return (s-c1)*(1/c2);
    }
  else
    {
    return 0.0;
    }
  }

float gen_rand(float ratio, float multiplier)
  {
  return (((float)suf_random())-ratio)*multiplier;
  }

float alt_mod(float alt, float mod)
  {
  return pow(fabs(alt),mod);
  }

int generate_fractal_terrain(float **t1, unsigned int *e1, vector ***w1, vector ***o1, float ***tp1, float ***tp2, float ***r1, char ***c1, int steps, float ratio, float ruggedness, int width, int height, int detail, int x_squares, int y_squares, float constant1, float constant2, float alti_mod, float mount_scale, float min_wind, float wind_scale, float ocean_scale, int av_size, float av_pow1, float av_pow2, float sun_scale, float rain_scale, float jungle_bonus, float swamp_bonus, float desert_bonus, float plains_bonus, float grassland_bonus, float forest_bonus, float hills_bonus, float tundra_bonus, float glacier_bonus, float mountain_bonus)
  {
  int a,b,c,x,y,x2,y2,x3,y3;
  long d,e;
  int seed;
  int x_dec,y_dec;
  int x_plus,y_plus;
  GLfloat az,bz,nx,ny,nz,r;

  vector **wind_map;
  vector **ocean_map;
  float **ocean_temp;
  float **atm_temp;
  float **rain_map;
  char **color_map;
  float ff,ff2,fg,fblee;
  float tw=width/((float)detail)*2;
  float th=height/((float)detail)*2;
  float f,f1,f2;
  
  float *terrain;
  float *terrain2;

  d=detail;
  e=d;
  
  seed=random();
  
  srandom(seed);

  vr_ratio=((float)detail)/((float)REAL_DETAIL);

  printf("Generating landscape...\n");

  if (!(terrain=malloc(e*e*sizeof(float))))
    {
    printf("Error(1): Unable to allocate enough memory for the terrain.\n");
    exit(1);
    }
  if (!(terrain2=malloc(e*e*sizeof(float))))
    {
    printf("Error(1): Unable to allocate enough memory for a terrain buffer.\n");
    exit(1);
    }

  fblee=sqrt(width*height)*mount_scale;
  terrain[detail*(0)+(0)]=gen_rand(ratio, fblee);
  terrain[detail*(0)+(e)]=gen_rand(ratio, fblee);
  for (a=1; a<=steps; a++)
    {
    ff=de_mod(a,steps,constant1,constant2);
    ff2=1-ff;
    for (x=0; x<e; x+=d)
      {
      if (x<e-d)
        x_dec=x-d;
      else
        x_dec=e-1;
      for (y=0; y<e; y+=d)
        {
        if (y<e-d)
          y_dec=y-d;
        else
          y_dec=e-1;
        terrain[detail*(x+d/2)+(y)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(x)+(y)]+terrain[detail*(x+d)+(y)],alti_mod)*d/e/2+(terrain[detail*(x)+(y)]+terrain[detail*(x+d)+(y)])/2)+gen_rand(ratio, fblee)*ff;
        terrain[detail*(x)+(y+d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(x)+(y)]+terrain[detail*(x)+(y+d)],alti_mod)*d/e/2+(terrain[detail*(x)+(y)]+terrain[detail*(x)+(y+d)])/2)+gen_rand(ratio, fblee)*ff;
        terrain[detail*(x+d/2)+(y+d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(x)+(y)]+terrain[detail*(x)+(y+d)]+terrain[detail*(x+d)+(y+d)]+terrain[detail*(x+d)+(y)],alti_mod)*d/e/4+(terrain[detail*(x)+(y)]+terrain[detail*(x)+(y+d)]+terrain[detail*(x+d)+(y+d)]+terrain[detail*(x+d)+(y)])/4)+gen_rand(ratio, fblee)*ff;
        }
      terrain[detail*(x+d/2)+(e-d)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(x)+(y)]+terrain[detail*(x+d)+(y)],alti_mod)*d/e/2+(terrain[detail*(x)+(y)]+terrain[detail*(x+d)+(y)])/2)+gen_rand(ratio, fblee)*ff;
      terrain[detail*(x)+(e-d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(x)+(y)]+terrain[detail*(x)+(0)],alti_mod)*d/e/2+(terrain[detail*(x)+(y)]+terrain[detail*(x)+(0)])/2)+gen_rand(ratio, fblee)*ff;
      terrain[detail*(x+d/2)+(e-d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(x)+(y)]+terrain[detail*(x)+(0)]+terrain[detail*(x+d)+(0)]+terrain[detail*(x+d)+(y)],alti_mod)*d/e/4+(terrain[detail*(x)+(y)]+terrain[detail*(x)+(0)]+terrain[detail*(x+d)+(0)]+terrain[detail*(x+d)+(y)])/4)+gen_rand(ratio, fblee)*ff;
      }
    for (y=0; y<e; y+=d)	//Wraparound
      {
      if (y<e-d)
        y_dec=y-d;
      else
        y_dec=e-1;
      terrain[detail*(e-d/2)+(y)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(e-d)+(y)]+terrain[detail*(0)+(y)],alti_mod)*d/e/2+(terrain[detail*(e-d)+(y)]+terrain[detail*(0)+(y)])/2)+gen_rand(ratio, fblee)*ff;
      terrain[detail*(e-d)+(y+d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(e-d)+(y)]+terrain[detail*(e-d)+(y+d)],alti_mod)*d/e/2+(terrain[detail*(e-d)+(y)]+terrain[detail*(e-d)+(y+d)])/2)+gen_rand(ratio, fblee)*ff;
      terrain[detail*(e-d/2)+(y+d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(e-d)+(y)]+terrain[detail*(e-d)+(y+d)]+terrain[detail*(0)+(y+d)]+terrain[detail*(0)+(y)],alti_mod)*d/e/4+(terrain[detail*(e-d)+(y)]+terrain[detail*(e)+(y+d)]+terrain[detail*(0)+(y+d)]+terrain[detail*(0)+(y)])/4)+gen_rand(ratio, fblee)*ff;
      }
    terrain[detail*(e-d/2)+(e-d)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(e-d)+(e-d)]+terrain[detail*(0)+(e-d)],alti_mod)*d/e/2+(terrain[detail*(e-d)+(e-d)]+terrain[detail*(0)+(e-d)])/2)+gen_rand(ratio, fblee)*ff;
    terrain[detail*(e-d)+(e-d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(e-d)+(e-d)]+terrain[detail*(e-d)+(0)],alti_mod)*d/e/2+(terrain[detail*(e-d)+(e-d)]+terrain[detail*(e-d)+(0)])/2)+gen_rand(ratio, fblee)*ff;
    terrain[detail*(e-d/2)+(e-d/2)]=ff2*(sf_random()*ruggedness*alt_mod(terrain[detail*(e-d)+(e-d)]+terrain[detail*(e-d)+(0)]+terrain[detail*(0)+(0)]+terrain[detail*(0)+(e-d)],alti_mod)*d/e/4+(terrain[detail*(e-d)+(e-d)]+terrain[detail*(e-d)+(0)]+terrain[detail*(0)+(0)]+terrain[detail*(0)+(e-d)])/4)+gen_rand(ratio, fblee)*ff;
    d/=2;
    }

  printf("Implementing erosion...\n");

  for (y=0; y<detail; y++)
    {
    for (x=0; x<detail; x++)
      {
      terrain2[x*detail+y]=0;
      f=0;
      for (a=0; a<av_size; a++)
        {
        for (y2=y-a; y2<=y+a; y2++)
          {
          if (y2<0)
            y3=y2+detail;
          else if (y2>=detail)
            y3=y2-detail;
          else
            y3=y2;
          for (x2=x-a; x2<=x+a; x2++)
            {
            if (x2<0)
              x3=x2+detail;
            else if (x2>=detail)
              x3=x2-detail;
            else
              x3=x2;
            f1=pow(1.0/(dist_2d(x,y,x3,y3)+1.0),av_pow1*terrain[x*detail+y]+av_pow2);
            terrain2[x*detail+y]+=terrain[x3*detail+y3]*f1;
            f+=f1;
            }
          }
        }
      terrain2[x*detail+y]=(terrain2[x*detail+y])/(f+0.00001);
      }
    }
   
   for (a=0; a<detail*detail; a++)
     terrain[a]=terrain2[a];

  free(terrain2);

  printf("Generating wind map...\n");

  wind_map=generate_wind_map(terrain, detail, min_wind, wind_scale, sqrt(width*height));

  printf("Generating ocean map...\n");

  ocean_map=generate_ocean_map(terrain, detail, ocean_scale, sqrt(width*height));

  printf("Determing ocean temperatures...\n");
  
  ocean_temp=generate_ocean_temp(ocean_map, detail, sun_scale);
  
  printf("Determining atmospheric temperatures...\n");

  atm_temp=generate_atm_temp(ocean_temp, wind_map, terrain, detail, sun_scale);
  
  printf("Determining rainfall...\n");
  
  rain_map=generate_rain_map(atm_temp, wind_map, terrain, detail, rain_scale);
  
  printf("Adjusting landscape to account for climate...\n");
  
  color_map=tweak_terrain(terrain, atm_temp, rain_map, detail, jungle_bonus, swamp_bonus, desert_bonus, plains_bonus, grassland_bonus, forest_bonus, hills_bonus, tundra_bonus, glacier_bonus, mountain_bonus);
  
  printf("Generating rivers and lakes...\n");
  
  printf("Adding resources...\n");
  
  printf("Building standard civ-compatable terrain map...\n");


  *t1=terrain;
  *e1=random();
  *w1=wind_map;
  *o1=ocean_map;
  *tp1=ocean_temp;
  *tp2=atm_temp;
  *r1=rain_map;
  *c1=color_map;

  if (!KEEP_WIND_AROUND)
    _2dfree((void **)wind_map,REAL_DETAIL);
  if (!KEEP_OCEAN_AROUND)
    _2dfree((void **)ocean_map,REAL_DETAIL);
  if (!KEEP_OTEMP_AROUND)
    _2dfree((void **)ocean_temp,REAL_DETAIL);
  if (!KEEP_ATEMP_AROUND)
    _2dfree((void **)atm_temp,REAL_DETAIL);
  if (!KEEP_RAIN_AROUND)
    _2dfree((void **)rain_map,REAL_DETAIL);

  printf("Launching... \n");

  return seed;
  }

float *zoom(float *terrain, char **colormap, int detail, int steps, float ruggedness, float alti_mod, int seed, int amount, int xloc, int yloc)
  {
  int a,b,x,y,x2,y2,x3,y3;
  float f,f1,f2;
  long d,e;
  int x_dec,y_dec;
  int x_plus,y_plus;
  float *terrain2;
//  float *terrain3;
  
  if (!(terrain2=malloc(detail*detail*sizeof(float))))
    {
    printf("Error(1): Unable to allocate enough memory to zoom.\n");
    exit(1);
    }
/*    
  if (!(terrain3=malloc(detail*detail*sizeof(float))))
    {
    printf("Error(2): Unable to allocate enough memory to zoom.\n");
    exit(1);
    }
*/
  b=1;
  for (a=0; a<amount; a++)
    b*=2;
  
//  printf("1) %d\n",xloc);
  
  if (xloc<detail/b/2)
    xloc=detail/b/2;
  if (xloc+detail/b>=detail)
    xloc=detail-1-detail/b/2;
  if ((yloc-detail/b/2)<0)
    yloc=detail/b/2;
  if (yloc+detail/b/2>=detail)
    yloc=detail-1-detail/b/2;

//  printf("2) %d\n",xloc);
  
  for (x=0; x<detail; x+=b)
    {
    for (y=0; y<detail; y+=b)
      {
      terrain2[x*detail+y]=terrain[(x/b+xloc-detail/2/b)*detail+(y/b+yloc-detail/2/b)]*b;
      }
    }

  srandom(seed);
  for (a=1; a<b; a++)
    {
    random();
    }

  d=detail;
  e=d;
  
  for (a=1; a<(steps-b)+2; a++)
    d/=2;
  
  for (a=(steps-b)+2; a<=steps; a++)
    {
    for (x=0; x<e; x+=d)
      {
      if (x<e-d)
        x_dec=x-d;
      else
        x_dec=e-1;
      for (y=0; y<e; y+=d)
        {
        if (y<e-d)
          y_dec=y-d;
        else
          y_dec=e-1;
        terrain2[detail*(x+d/2)+(y)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(x)+(y)]+terrain2[detail*(x+d)+(y)],alti_mod)*d/e/2+(terrain2[detail*(x)+(y)]+terrain2[detail*(x+d)+(y)])/2);
        terrain2[detail*(x)+(y+d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(y+d)],alti_mod)*d/e/2+(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(y+d)])/2);
        terrain2[detail*(x+d/2)+(y+d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(y+d)]+terrain2[detail*(x+d)+(y+d)]+terrain2[detail*(x+d)+(y)],alti_mod)*d/e/4+(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(y+d)]+terrain2[detail*(x+d)+(y+d)]+terrain2[detail*(x+d)+(y)])/4);
        }
      terrain2[detail*(x+d/2)+(e-d)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(x)+(y)]+terrain2[detail*(x+d)+(y)],alti_mod)*d/e/2+(terrain2[detail*(x)+(y)]+terrain2[detail*(x+d)+(y)])/2);
      terrain2[detail*(x)+(e-d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(0)],alti_mod)*d/e/2+(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(0)])/2);
      terrain2[detail*(x+d/2)+(e-d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(0)]+terrain2[detail*(x+d)+(0)]+terrain2[detail*(x+d)+(y)],alti_mod)*d/e/4+(terrain2[detail*(x)+(y)]+terrain2[detail*(x)+(0)]+terrain2[detail*(x+d)+(0)]+terrain2[detail*(x+d)+(y)])/4);
      }
    for (y=0; y<e; y+=d)	//Wraparound
      {
      if (y<e-d)
        y_dec=y-d;
      else
        y_dec=e-1;
      terrain2[detail*(e-d/2)+(y)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(e-d)+(y)]+terrain2[detail*(0)+(y)],alti_mod)*d/e/2+(terrain2[detail*(e-d)+(y)]+terrain2[detail*(0)+(y)])/2);
      terrain2[detail*(e-d)+(y+d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(e-d)+(y)]+terrain2[detail*(e-d)+(y+d)],alti_mod)*d/e/2+(terrain2[detail*(e-d)+(y)]+terrain2[detail*(e-d)+(y+d)])/2);
      terrain2[detail*(e-d/2)+(y+d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(e-d)+(y)]+terrain2[detail*(e-d)+(y+d)]+terrain2[detail*(0)+(y+d)]+terrain2[detail*(0)+(y)],alti_mod)*d/e/4+(terrain2[detail*(e-d)+(y)]+terrain2[detail*(e)+(y+d)]+terrain2[detail*(0)+(y+d)]+terrain2[detail*(0)+(y)])/4);
      }
    terrain2[detail*(e-d/2)+(e-d)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(e-d)+(e-d)]+terrain2[detail*(0)+(e-d)],alti_mod)*d/e/2+(terrain2[detail*(e-d)+(e-d)]+terrain2[detail*(0)+(e-d)])/2);
    terrain2[detail*(e-d)+(e-d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(e-d)+(e-d)]+terrain2[detail*(e-d)+(0)],alti_mod)*d/e/2+(terrain2[detail*(e-d)+(e-d)]+terrain2[detail*(e-d)+(0)])/2);
    terrain2[detail*(e-d/2)+(e-d/2)]=(sf_random()*ruggedness*alt_mod(terrain2[detail*(e-d)+(e-d)]+terrain2[detail*(e-d)+(0)]+terrain2[detail*(0)+(0)]+terrain2[detail*(0)+(e-d)],alti_mod)*d/e/4+(terrain2[detail*(e-d)+(e-d)]+terrain2[detail*(e-d)+(0)]+terrain2[detail*(0)+(0)]+terrain2[detail*(0)+(e-d)])/4);
    d/=2;
    }

  //We could put special cases here to handle the zooms of different terrains differently... erode some at certain levels, make them jagged or rounded at others, etc.... but a) I'm too lazy, but more importantly, b) I don;t have time
  //Hehehe... I'm speaking in the queen's english  ;)

/*
  for (y=0; y<detail; y++)
    {
    for (x=0; x<detail; x++)
      {
      terrain3[x*detail+y]=0;
      f=0;
      for (a=0; a<1; a++)
        {
        for (y2=y-a; y2<=y+a; y2++)
          {
          if (y2<0)
            y3=y2+detail;
          else if (y2>=detail)
            y3=y2-detail;
          else
            y3=y2;
          for (x2=x-a; x2<=x+a; x2++)
            {
            if (x2<0)
              x3=x2+detail;
            else if (x2>=detail)
              x3=x2-detail;
            else
              x3=x2;
            f1=pow(1.0/(dist_2d(x,y,x3,y3)+1.0),0.05*terrain2[x*detail+y]);
            terrain3[x*detail+y]+=terrain2[x3*detail+y3]*f1;
            f+=f1;
            }
          }
        }
      terrain3[x*detail+y]=(terrain3[x*detail+y])/(f+0.00001);
      }
    }
    
  for (a=0; a<detail*detail; a++)
    terrain2[a]=terrain3[a];
  
  free(terrain3);
*/  
  return terrain2;
  }

vector find_downslope(float *terrain, int x, int y, int detail, float scale)
  {
  vector tmp;
  float xc,yc,slope;
  int a;
  float tmp2=terrain[x*detail+y];
  float z[9];
  int x_plus=x+1;
  int y_plus=y+1;
  int x_min=x-1;
  int y_min=y-1;
  if (x_plus==detail)
    x_plus-=detail;
  else if (x_min<0)
    x_min+=detail;
  if (y_plus==detail)
    y_plus-=detail;
  else if (y_min<0)
    y_min+=detail;
  if (tmp2<0)
    tmp2=0;
  z[0]=terrain[x_min*detail+y_min];
  z[1]=terrain[x*detail+y_min];
  z[2]=terrain[x_plus*detail+y_min];
  z[3]=terrain[x_min*detail+y];
  z[5]=terrain[x_plus*detail+y];
  z[6]=terrain[x_min*detail+y_plus];
  z[7]=terrain[x*detail+y_plus];
  z[8]=terrain[x_plus*detail+y_plus];
  for (a=0; a<9; a++)
    {
    if (z[a]<0)	//If it's ocean.  Once we have rivers and lakes, we'll add them in too.
      z[a]=-tmp2;
    else
      z[a]-=tmp2;
    }
  xc=z[5]*.3-z[3]*.3+z[8]*.1+z[2]*.1-z[0]*.1-z[6]*.1;
  yc=z[7]*.3-z[1]*.3+z[8]*.1+z[6]*.1-z[0]*.1-z[2]*.1;
  if (xc==0)
    {
    if (yc<0)
      {
      tmp.angle=PI;
      }
    else
      {
      tmp.angle=0;
      }
    }
  else
    {
    slope=yc/xc;
    if (xc>0)
      tmp.angle=PI*3/2-atan(slope);
    else
      tmp.angle=PI*1/2-atan(slope);
    }
  tmp.speed=sqrt(xc*xc+yc*yc)*tmp2/scale/2;
//  printf("%f %f %f %f\n",tmp.angle,tmp.speed,tmp2,terrain[x*detail+y]);
  return tmp;
  }

vector ang_average(float ang1, float speed1, float ang2, float speed2)
  {
  vector newang;
  speed1=fabs(speed1);
  speed2=fabs(speed2);
  newang.speed=sqrt(speed1*speed1+speed2*speed2)/(speed1+speed2);
  if (fabs(ang1-ang2)<PI)
    newang.angle=(ang1*speed1+ang2*speed2)/(speed1+speed2);
  else
    {
    if (ang1<ang2)
      ang1+=2*PI;
    else
      ang2+=2*PI;
    if ((newang.angle=(ang1*speed1+ang2*speed2)/(speed1+speed2))>2*PI)
      newang.angle-=2*PI;
    }
  
//  printf("%f %f + %f %f = %f %f\n",ang1,speed1,ang2,speed2,newang.angle,newang.speed);
  return newang;
  }

float ang_diff(float ang1, float ang2)
  {
  if (fabs(ang1-ang2)>PI)
    return 2*PI-fabs(ang1-ang2);
  else
    return fabs(ang1-ang2); 
  }

vector **generate_wind_map(float *terrain, int detail, float min_wind, float wind_scale, float cnst)	//Sorry, the jet streams and el nino aren't taken into effect.
  {
  int a,x,y;
  float fd=(float)REAL_DETAIL;
  float tmp,tmp2;
  vector **windp;
//  float **newspeed;
  float speedav;
  vector grad;
  vector new;
  if (!(windp=malloc(REAL_DETAIL*sizeof(vector *))))
    {
    printf("Error(1): Unable to allocate enough memory for the wind map.\n");
    exit(1);
    }
/*
  if (!(newspeed=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the newspeed map.\n");
    exit(1);
    }
*/
  for (a=0; a<=detail; a++)
    {
    if (!(windp[a]=malloc(REAL_DETAIL*sizeof(vector))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the wind map.\n",a);
      exit(1);
      }
/*
    if (!(newspeed[a]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the newspeed map.\n",a);
      exit(1);
      }
*/
    }
  //Ok, I'll cheat a little on the polar highs and incorporate them into the polar easterlies.. so sue me  :)  They actually rotate clockwise (like the earth)
  //The highs will be treated as nondirectional, full speed, since in rl they create spirals and storm systems randomly.  The highs will be treated as nondirectional no speed, as they don't have storms themselves.  The intertropical zone will be nondirectional, nonspeed (too varied)
  //BTW, I don't have lattitude ranges on the start of the lows and highs, so I'm guestimating that they're each 5 degrees longitude wide.
  //Oh, and I'm also guessing about the windspeeds of the various currents, assuming that the horizontal currents are the fastest due to earth's rotation.

  printf("   implementing coriolis effect.\n");

  //North polar high and polar easterlies
  tmp=1/fd*180/27.5;
  for (y=0; y<fd*27.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=PI+y*tmp*PI/2;
      windp[x][y].speed=y*tmp2+min_wind;
      }
    }
  // Subpolar low
  for (y=fd*27.5/180; y<fd*32.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=-1;
      windp[x][y].speed=tmp2+min_wind;
      }
    }
  //Westerlies
  for (y=fd*32.5/180; y<fd*57.5/180; y++)
    {
    tmp=1/fd*180/25;
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=0+(y-fd*32.5/180)*tmp*PI/2;
      windp[x][y].speed=1.0-(y-fd*32.5/180)*tmp+min_wind;
      }
    }
  // Subtropical high (horse latitudes)
  for (y=fd*57.5/180; y<fd*62.5/180; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=-1;
      windp[x][y].speed=min_wind;
      }
    }
  //NE trade winds
  for (y=fd*62.5/180; y<fd*87.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=PI/2+(y-fd*62.5/180)*tmp*PI/2;
      windp[x][y].speed=(y-fd*62.5/180)*tmp+min_wind;
      }
    }
  //Intertropical convergence zone (doldrums)
  for (y=fd*87.5/180; y<fd*92.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=-1;
      windp[x][y].speed=tmp2+min_wind;
      }
    }
  //SE trade winds
  for (y=fd*92.5/180; y<fd*117.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=PI/2-(y-fd*92.5/180)*tmp*PI/2;
      windp[x][y].speed=1.0-(y-fd*92.5/180)*tmp+min_wind;
      }
    }
  //Subtropical high (horse latitudes)
  for (y=fd*117.5/180; y<fd*122.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=-1;
      windp[x][y].speed=tmp2+min_wind;
      }
    }
  //Westerlies
  for (y=fd*122.5/180; y<fd*157.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=PI*2-(y-fd*122.5/180)*tmp*PI/2;
      windp[x][y].speed=(y-fd*122.5/180)*tmp+min_wind;
      }
    }
  //Subpolar low
  for (y=fd*157.5/180; y<fd*162.5/180; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=-1;
      windp[x][y].speed=tmp2+min_wind;
      }
    }
  //South polar easterlies and polar high
  for (y=fd*162.5/180; y<fd; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=PI/2-(y-fd*162.5/180)*tmp*PI/2;
      windp[x][y].speed=1-(y-fd*162.5/180)*tmp+min_wind;
      }
    }
/*  Debugging wind
  for (y=0; y<fd; y++)
    {
    tmp2=tmp*sqrt(1-y/fd);
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].angle=PI*5/8;
      windp[x][y].speed=y*tmp2+min_wind;
      }
    }
*/
  //Now, we need to diverge around the landscape
  printf("   diverging around landscape.\n");

  for (y=0; y<REAL_DETAIL; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
//      printf("%d %d) ",x,y);
      grad=find_downslope(terrain, x*vr_ratio, y*vr_ratio, detail, cnst);
      new=ang_average(windp[x][y].angle,windp[x][y].speed,grad.angle,grad.speed);
      windp[x][y].angle=new.angle;
//      newspeed[x][y]=new.speed;
      windp[x][y].speed=new.speed;
      }
    }
  for (x=0; x<REAL_DETAIL; x++)
    {
    for (y=0; y<REAL_DETAIL; y++)
      {
      windp[x][y].speed*=wind_scale;
      }
    }
/*
  speedav=0;			//Find the average speed increase
  for (y=0; y<REAL_DETAIL; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
      speedav+=newspeed[x][y]/windp[x][y].speed;
      }
    }
  speedav/=(fd*fd);		//Now, scale them down so it averages out to zero change
  for (y=0; y<REAL_DETAIL; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
      windp[x][y].speed=newspeed[x][y]/speedav;
      }
    }
  free(newspeed);
*/
  return windp;
  }

static float current_angle(int x, int y, float *terrain, int detail)
  {
  //Send out N rays, and find how much resistance they hit.  Determine your
  //angle from that.
  float resistance;
  float angle;
  float x2,y2,x3,y3,f,f1,f2,borg,slope;
  int n,a,b,c;
  x3=0;
  y3=0;
  if (terrain[x*detail+y]>0)
    return FUTILE;
  for (n=0; n<OCEAN_ANGLE_COUNT; n++)
    {
    angle=2*PI*n/OCEAN_ANGLE_COUNT;
    f1=sin(angle);
    f2=cos(angle);
    x2=x+0.5;
    y2=y+0.5;
    resistance=0;
    for (a=0; a<(int)(OCEAN_ANGLE_COUNT*((1+WESTWARD_INTENSIFICATION)+f1*WESTWARD_INTENSIFICATION)); a++)
      {
      x2+=f1;
      y2+=f2;
      if (x2>=detail)
        x2-=detail;
      else if (x2<0)
        x2+=detail;
      if (y2>=detail)
        y2-=detail;
      else if (y2<0)
        y2+=detail;
      
      borg=terrain[((int)x2)*detail+((int)y2)];
      if (borg<0)
        {
        f=1.0/((-borg+30.0)*(((float)a)+1.0*((1+WESTWARD_INTENSIFICATION)+f1*WESTWARD_INTENSIFICATION)));
        resistance+=f;
        }
      else
        {
        resistance+=1/(30.0*(((float)a)+1.0*((1+WESTWARD_INTENSIFICATION)+f1*WESTWARD_INTENSIFICATION)));
        }
      }
    x3+=f1*resistance;
    y3+=f2*resistance;
//    printf("%f %f %f %f %f\n",x3,y3,f1,f2,resistance);
    }
  if (x3==0)
    {
    if (y3<0)
      {
      return PI;
      }
    else
      {
      return 0;
      }
    }
  else
    {
    slope=y3/x3;
    if (x3>0)				//Yes, the angle is rotated 180 degrees
      return PI*1/2-atan(slope);
    else
      return PI*3/2-atan(slope);
    }
  }

static float current_speed(int x, int y, float *terrain, vector **ocean, int detail)
  {
  //Trace backwards along the angle, checking shallowness along the way.
  //Trace back N steps, and have a total water mass accumulated along the way,
  //relative to how from the starting position we are.  Hmm... we should be
  //nice and branch at each step, and simply kill paths that don't produce
  //enough water, but never stopping before N.

  int branchx[OCEAN_SPEED_MAXBR];
  int branchy[OCEAN_SPEED_MAXBR];
  float branchs[OCEAN_SPEED_MAXBR];
  float old_ang[OCEAN_SPEED_MAXBR];
  
  int n,a,b,branches,new_branches,bstart,xc,yc;
  float sum=0.0;
  float tmp,opp_ang;
  
  branches=1;
  bstart=0;
  branchx[0]=x;
  branchy[0]=y;
  branchs[0]=1.0;
  opp_ang=PI+ocean[x][y].angle;
  if (opp_ang>=2*PI)
    opp_ang-=2*PI;
  old_ang[0]=opp_ang;

  for (n=0; n<OCEAN_SPEED_COUNT; n++)
    {
    //Run through each branch, add its volume to the sum, see if it needs to
    //be eliminated or branch itself.
    a=bstart;
    new_branches=0;
    for ( ; a<bstart+branches; a++)
      {
      //First, add the volume, relative to how far into it we are.
      tmp=-terrain[((int)(branchx[a]*vr_ratio))*detail+((int)(branchy[a]*vr_ratio))];
      if (tmp<0)
        tmp=0;
      tmp=sqrt(tmp);
      opp_ang=PI+ocean[branchx[a]][branchy[a]].angle;
      if (opp_ang>=2*PI)
        opp_ang-=2*PI;
      tmp/=(ang_diff(opp_ang,old_ang[a])+0.0001);
      sum+=tmp*branchs[a];
      //Now, is that amount large enough to branch or does it just die?
      if (tmp>1)	//branch
        {
        xc=branchx[a]+(opp_ang<PI?1:-1);
        if (xc<0)
          xc+=REAL_DETAIL;
        else if (xc>=REAL_DETAIL)
          xc-=REAL_DETAIL;
        branchx[bstart+branches+new_branches]=xc;	//branch 1
        branchy[bstart+branches+new_branches]=branchy[a];
        branchs[bstart+branches+new_branches]=fabs(sin(opp_ang))*branchs[a];
        old_ang[bstart+branches+new_branches]=opp_ang;
        for (b=0; b<bstart; b++)
          {
          if ((branchx[b]==branchx[bstart+branches+new_branches])&&(branchy[b]==branchy[bstart+branches+new_branches]))
            {
            if (branchs[b]==1.0)
              break;
            else
              {
              if (branchs[b]+branchs[a]>=1.0)
                branchs[a]=1-branchs[b];
//              else
//                branchs[a]=branchs[a]+branchs[b];
              branchs[b]=1.0;
              }
            }
          }
        if (b==bstart)		//If its not a duplicate
          {
          new_branches++;
          if (bstart+branches+new_branches>=OCEAN_SPEED_MAXBR)
            return sqrt(sum)/OCEAN_SPEED_DIVISOR;
          }
        yc=branchy[a]+((opp_ang<PI/2)||(opp_ang>PI*3/2)?-1:1);
        if (yc<0)
          yc=0;
        else if (yc>=REAL_DETAIL)
          yc=REAL_DETAIL;
        branchx[bstart+branches+new_branches]=branchx[a];	//branch 2
        branchy[bstart+branches+new_branches]=yc;
        branchs[bstart+branches+new_branches]=fabs(cos(opp_ang))*branchs[a];
        old_ang[bstart+branches+new_branches]=opp_ang;
        for (b=0; b<bstart; b++)
          {
          if ((branchx[b]==branchx[bstart+branches+new_branches])&&(branchy[b]==branchy[bstart+branches+new_branches]))
            {
            if (branchs[b]==1.0)
              break;
            else
              {
              if (branchs[b]+branchs[a]>=1.0)
                branchs[a]=1-branchs[b];
//              else
//                branchs[a]=branchs[a]+branchs[b];
              branchs[b]=1.0;
              }
            }
          }
        if (b==bstart)		//Again, if not a duplicate
          {
          new_branches++;
          if (bstart+branches+new_branches>=OCEAN_SPEED_MAXBR)
            return sqrt(sum)/OCEAN_SPEED_DIVISOR;
          }
        }
      }
    bstart=bstart+branches;
    branches=new_branches;
    }
  return sqrt(sum)/OCEAN_SPEED_DIVISOR;
  }

vector **generate_ocean_map(float *terrain, int detail, float ocean_scale, float cnst)
  {
  int a,b,x,y;
  float fd=(float)REAL_DETAIL;
  float borg,f;
  float resistance;
  vector **ocean;
  if (!(ocean=malloc(REAL_DETAIL*sizeof(vector *))))
    {
    printf("Error(1): Unable to allocate enough memory for the ocean map.\n");
    exit(1);
    }
  for (a=0; a<=REAL_DETAIL; a++)
    {
    if (!(ocean[a]=malloc(REAL_DETAIL*sizeof(vector))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the ocean map.\n",a);
      exit(1);
      }
    }
/*    Old Code:
      borg=floor_height_in(x,y,detail,ocean[x][y].angle);
      if (borg!=0)
        resistance=terrain[x*detail+y]/borg;
      else
        resistance=FUTILE;	//Infinite  (ok, so I had to stick a little joke in there  ;)  Futile is just a constant, means -1
      }
    }
*/
  printf("   calculating ocean angles.\n");
  for (y=0; y<REAL_DETAIL; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
      ocean[x][y].angle=current_angle(x,y,terrain,detail);
      if (y>=REAL_DETAIL/2)
        {
        ocean[x][y].angle+=PI/2;
        if (ocean[x][y].angle>=PI*2)
          ocean[x][y].angle-=PI*2;
        }
      else
        {
        ocean[x][y].angle-=PI/2;
        if (ocean[x][y].angle<0)
          ocean[x][y].angle+=PI*2;
        }
      }
    }
  printf("   calculating ocean speeds.\n");
  f=MOUNTAIN_SCALAR;
  for (y=0; y<REAL_DETAIL; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
      ocean[x][y].speed=current_speed(x,y,terrain,ocean,detail)*ocean_scale*f;
      }
    }
  return ocean;
  }

float **generate_ocean_temp(vector **ocean, int detail, float scale)
  {
  int a,x,y,x_dec,x_inc,y_dec,y_inc;
  float f,speed,f1,f2;
  float **otemp;
  float **otemp2;
  float av_temp;
  if (!(otemp=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the ocean temperature map.\n");
    exit(1);
    }
  if (!(otemp2=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the ocean temperature map2.\n");
    exit(1);
    }
  for (a=0; a<=REAL_DETAIL; a++)
    {
    if (!(otemp[a]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the ocean temperature map.\n",a);
      exit(1);
      }
    if (!(otemp2[a]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the ocean temperature map2.\n",a);
      exit(1);
      }
    }
  f=scale/MOUNTAIN_SCALAR*4;
  printf("   initializing.\n");
  for (y=0; y<REAL_DETAIL/2+1; y++)
    {		//We'll assume solar warming is linear (in kelvins)
    av_temp=f*y*NATURAL_TEMPERATURE_RANGE+NATURAL_TEMPERATURE_BASE;
    for (x=0; x<REAL_DETAIL; x++)
      {
      otemp[x][y]=av_temp;
      otemp[x][REAL_DETAIL-y]=av_temp;
      }
    }
  printf("   mixing.\n");
  for (a=0; a<OCEAN_MIXING_PASSES; a++)
    {
    for (y=0; y<REAL_DETAIL; y++)
      {
      if (y==0)
        {
        y_dec=REAL_DETAIL-1;
        y_inc=1;
        }
      else if (y==REAL_DETAIL-1)
        {
        y_dec=REAL_DETAIL-2;
        y_inc=0;
        }
      else
        {
        y_dec=y-1;
        y_inc=y+1;
        }
      for (x=0; x<REAL_DETAIL; x++)
        {
        if (x==0)
          {
          x_dec=REAL_DETAIL-1;
          x_inc=1;
          }
        else if (x==REAL_DETAIL-1)
          {
          x_dec=REAL_DETAIL-2;
          x_inc=0;
          }
        else
          {
          x_dec=x-1;
          x_inc=x+1;
          }
        speed=ocean[x][y].speed;
        f1=O_HEAT_DISTRIBUTION_CONSTANT/(speed+0.0001);
        f2=speed/(f1+speed);
        f1=f1/(f1+speed);
//        printf("%f %f\n",f1,f2);
        otemp2[x][y]=otemp[x][y]*f1+(( (ocean[x][y].angle>PI) ? otemp[x_dec][y]*square(sin(ocean[x][y].angle+PI)) : otemp[x_inc][y]*square(sin(ocean[x][y].angle+PI)))+( ((ocean[x][y].angle<PI/2)||(ocean[x][y].angle>PI*3/2)) ? otemp[x][y_dec]*square(cos(ocean[x][y].angle+PI)) : otemp[x][y_inc]*square(cos(ocean[x][y].angle+PI))))*f2; 
        }
      }
    for (y=0; y<REAL_DETAIL; y++)
      {
      for (x=0; x<REAL_DETAIL; x++)
        {
        otemp[x][y]=otemp2[x][y];
        }
      }
    }
  _2dfree((void **)otemp2,REAL_DETAIL);
  return otemp;
  }

float **generate_atm_temp(float **otemp, vector **wind, float *terrain, int detail, float scale)
  {
  int a,x,y,x_dec,x_inc,y_dec,y_inc;	//Cut 'n paste, change 'a bit  ;)
  float f,speed,f1,f2;			//pass it air, and ocean temp, not ocean!  :)
  float **atemp;
  float **atemp2;
  float av_temp;
  if (!(atemp=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the ocean temperature map.\n");
    exit(1);
    }
  if (!(atemp2=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the ocean temperature map2.\n");
    exit(1);
    }
  for (a=0; a<=REAL_DETAIL; a++)
    {
    if (!(atemp[a]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the ocean temperature map.\n",a);
      exit(1);
      }
    if (!(atemp2[a]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the ocean temperature map2.\n",a);
      exit(1);
      }
    }
  f=scale/MOUNTAIN_SCALAR*4;
  printf("   initializing.\n");
  for (y=0; y<REAL_DETAIL/2+1; y++)
    {		//We'll assume solar warming is linear (in kelvins)
    av_temp=f*y*NATURAL_TEMPERATURE_RANGE+NATURAL_TEMPERATURE_BASE;
    for (x=0; x<REAL_DETAIL; x++)
      {
      atemp[x][y]=av_temp;
      atemp[x][REAL_DETAIL-y]=av_temp;
      }
    }
  printf("   terrain adjusting.\n");
  f=HEIGHT_HEAT_CONST;
  for (y=0; y<REAL_DETAIL; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
//      printf("%f %f %f\n",terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))],atemp[x][y],f/(f+terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))]));
      if (terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))]>=0)
        atemp[x][y]=atemp[x][y]*f/(f+terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))]);
      else
        atemp[x][y]=atemp[x][y]*(1-OCEAN_HEAT_TRANSFER)+otemp[x][y]*OCEAN_HEAT_TRANSFER;
      }
    }
  printf("   mixing.\n");
  for (a=0; a<ATM_MIXING_PASSES; a++)
    {
    for (y=0; y<REAL_DETAIL; y++)
      {
      if (y==0)
        {
        y_dec=REAL_DETAIL-1;
        y_inc=1;
        }
      else if (y==REAL_DETAIL-1)
        {
        y_dec=REAL_DETAIL-2;
        y_inc=0;
        }
      else
        {
        y_dec=y-1;
        y_inc=y+1;
        }
      for (x=0; x<REAL_DETAIL; x++)
        {
        if (x==0)
          {
          x_dec=REAL_DETAIL-1;
          x_inc=1;
          }
        else if (x==REAL_DETAIL-1)
          {
          x_dec=REAL_DETAIL-2;
          x_inc=0;
          }
        else
          {
          x_dec=x-1;
          x_inc=x+1;
          }
        speed=wind[x][y].speed;
        f1=A_HEAT_DISTRIBUTION_CONSTANT/(speed+0.0001);
        f2=speed/(f1+speed);
        f1=f1/(f1+speed);
//        printf("%f %f\n",f1,f2);
        atemp2[x][y]=atemp[x][y]*f1+(( (wind[x][y].angle>PI) ? atemp[x_dec][y]*square(sin(wind[x][y].angle+PI)) : atemp[x_inc][y]*square(sin(wind[x][y].angle+PI)))+( ((wind[x][y].angle<PI/2)||(wind[x][y].angle>PI*3/2)) ? atemp[x][y_dec]*square(cos(wind[x][y].angle+PI)) : atemp[x][y_inc]*square(cos(wind[x][y].angle+PI))))*f2; 
        }
      }
    for (y=0; y<REAL_DETAIL; y++)
      {
      for (x=0; x<REAL_DETAIL; x++)
        {
        atemp[x][y]=atemp2[x][y];
//        printf("%f\n",atemp[x][y]);
        }
      }
    }

  _2dfree((void **)atemp2,REAL_DETAIL);
  return atemp;
  }

float **generate_rain_map(float **atemp, vector **wind, float *terrain, int detail, float scale)
  {
  int a,x,y,x_dec,x_inc,y_dec,y_inc;	//Cut 'n paste, change 'a bit  ;)
  float f,speed,f1,f2;			//pass it air, and ocean temp, not ocean!  :)
  float **rain;
  float **rain2;
  if (!(rain=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the rain map.\n");
    exit(1);
    }
  if (!(rain2=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the rain map2.\n");
    exit(1);
    }
  for (a=0; a<=REAL_DETAIL; a++)
    {
    if (!(rain[a]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the rain map.\n",a);
      exit(1);
      }
    if (!(rain2[a]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the rain map2.\n",a);
      exit(1);
      }
    }
  f=scale/MOUNTAIN_SCALAR*4;
  for (a=0; a<RAIN_MIXING_PASSES; a++)
    {
    for (y=0; y<REAL_DETAIL; y++)
      {
      if (y==0)
        {
        y_dec=REAL_DETAIL-1;
        y_inc=1;
        }
      else if (y==REAL_DETAIL-1)
        {
        y_dec=REAL_DETAIL-2;
        y_inc=0;
        }
      else
        {
        y_dec=y-1;
        y_inc=y+1;
        }
      for (x=0; x<REAL_DETAIL; x++)
        {
        if (x==0)
          {
          x_dec=REAL_DETAIL-1;
          x_inc=1;
          }
        else if (x==REAL_DETAIL-1)
          {
          x_dec=REAL_DETAIL-2;
          x_inc=0;
          }
        else
          {
          x_dec=x-1;
          x_inc=x+1;
          }
        if (terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))]<=0)
          {
          rain[x][y]+=atemp[x][y]*DEWPOINT_CONSTANT*scale;
          rain[x][y]-=pow(RAIN_PER_CYCLE,rain[x][y]/RAIN_NORMAL_FALL+RAIN_MIN_PERCENT);
          }
        else
          {
          rain[x][y]+=0;
          rain[x][y]-=pow(RAIN_PER_CYCLE,rain[x][y]/RAIN_NORMAL_FALL+RAIN_MIN_PERCENT);
          }
        speed=wind[x][y].speed;
        f1=RAIN_DISTRIBUTION_CONSTANT/(speed+0.0001);
        f2=speed/(f1+speed);
        f1=f1/(f1+speed);
//        printf("%f %f\n",f1,f2);
        rain2[x][y]=rain[x][y]*f1+(( (wind[x][y].angle>PI) ? rain[x_dec][y]*square(sin(wind[x][y].angle+PI)) : rain[x_inc][y]*square(sin(wind[x][y].angle+PI)))+( ((wind[x][y].angle<PI/2)||(wind[x][y].angle>PI*3/2)) ? rain[x][y_dec]*square(cos(wind[x][y].angle+PI)) : rain[x][y_inc]*square(cos(wind[x][y].angle+PI))))*f2; 
        }
      }
    for (y=0; y<REAL_DETAIL; y++)
      {
      for (x=0; x<REAL_DETAIL; x++)
        {
        rain[x][y]=rain2[x][y];
        }
      }
    }

  for (y=0; y<REAL_DETAIL; y++)
    {
    for (x=0; x<REAL_DETAIL; x++)
      {
      rain[x][y]=pow(RAIN_PER_CYCLE,rain[x][y]/RAIN_NORMAL_FALL+RAIN_MIN_PERCENT);
      }
    }

  _2dfree((void **)rain2,REAL_DETAIL);
  return atemp;
  }

int greaterthan(const void *a, const void *b)
  {
  float ax,bx;
  ax=*((float *)a);
  bx=*((float *)b);
  if (ax>=bx)
    {
    if (ax>bx)
      return 1;
    else
      return 0;
    }
  else
    return -1;
  }

float **generate_rugged_map(float *terrain, float **rugged, int detail)
  {
  int x,y,a,nx,ny,nx2,ny2;
  float av_height;
  for (x=0; x<REAL_DETAIL; x++)
    {
    for (y=0; y<REAL_DETAIL; y++)
      {
      for (nx=x-ROUGHNESS_SIZE; nx<=ROUGHNESS_SIZE; nx++)
        {
        if (nx<0)
          nx2=nx+REAL_DETAIL;
        else if (nx>REAL_DETAIL-1)
          nx2-nx-REAL_DETAIL;
        else
          nx2=nx;
        for (ny=y-ROUGHNESS_SIZE; ny<=ROUGHNESS_SIZE; ny++)
          {
          if (ny<0)
            ny2=ny+REAL_DETAIL;
          else if (ny>REAL_DETAIL-1)
            ny2-ny-REAL_DETAIL;
          else
            ny2=ny;
          av_height+=terrain[((int)(nx2*vr_ratio))*detail+((int)(ny2*vr_ratio))];
          }
        }
      av_height/=((ROUGHNESS_SIZE*2+1)*(ROUGHNESS_SIZE*2+1));
      rugged[x][y]=(terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))]-av_height)/MOUNTAIN_SCALAR;
      }
    }
  return rugged;
  }

char **tweak_terrain(float *terrain, float **atemp, float **rain, int detail, float jungle_bonus, float swamp_bonus, float desert_bonus, float plains_bonus, float grassland_bonus, float forest_bonus, float hills_bonus, float tundra_bonus, float glacier_bonus, float mountain_bonus)
  {		//Make the landscape match its surrounding conditions.  Also, create an unfinished colormap.
  int x,y,i,x_inc,x_dec,y_inc,y_dec;
  float r,t,u,a,f,x2,y2;
  char v[50];		//10 floats, 10 chars
  float **rugged;
  char **colormap;

  if (!(colormap=malloc(REAL_DETAIL*sizeof(char *))))
    {
    printf("Error(1): Unable to allocate enough memory for the colormap.\n");
    exit(1);
    }
  if (!(rugged=malloc(REAL_DETAIL*sizeof(float *))))
    {
    printf("Error(1): Unable to allocate enough memory for the colormap.\n");
    exit(1);
    }

  if (JUNGLE_MOUNTAINS_ONLY)
    return colormap;

  for (i=0; i<=REAL_DETAIL; i++)
    {
    if (!(colormap[i]=malloc(REAL_DETAIL*sizeof(char)+20)))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the colormap.\n",i);
      exit(1);
      }
    if (!(rugged[i]=malloc(REAL_DETAIL*sizeof(float))))
      {
      printf("Error(2,%d): Unable to allocate enough memory for the rugged.\n",i);
      exit(1);
      }
    }

  rugged=generate_rugged_map(terrain,rugged,detail);
  
/*	Official definitions:
  wet equatorial				150-1000, ~312
  tropical monsoon				Same, but seasonal
  tropical wet-dry				50-175,  ~305, more varied
  tropical-subtropical desert			0-25, 288+
  tropical-substropical steppe			38-63, same as above
  mid-latitude steppe and desert		Same, as above 2, cooler
  humid subtropical climate			75-150, 278-305
  mediterranean climate				30-90, same as above assumedly
  marine west coast climate			50-500, 273-293
  humid continental				50-125, same, more varied
  continental subarctic				<50, 223-333
  tundra climate				25-100,	cold-283
  snow and ice climate				5-50, 213-cold
  highland climates				no details
*/

  for (y=0; y<REAL_DETAIL; y++)
    {
    if (y==REAL_DETAIL-1)
      {
      y_inc=0;
      y_dec=REAL_DETAIL-2;
      }
    else if (y==0)
      {
      y_inc=1;
      y_dec=REAL_DETAIL-1;
      }
    else
      {
      y_inc=y+1;
      y_dec=y-1;
      }

    for (x=0; x<REAL_DETAIL; x++)
      {
      if (x==REAL_DETAIL-1)
        {
        x_inc=0;
        x_dec=REAL_DETAIL-2;
        }
      else if (x==0)
        {
        x_inc=1;
        x_dec=REAL_DETAIL-1;
        }
      else
        {
        x_inc=x+1;
        x_dec=x-1;
        }

      a=terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))]/MOUNTAIN_SCALAR/1.5;
      if (a<=0)
        {
        colormap[x][y]=2;
        continue;
        }
      a=2*pow(a,0.125);
      r=2*pow(rain[x][y]/(DEWPOINT_CONSTANT*(NATURAL_TEMPERATURE_BASE+NATURAL_TEMPERATURE_RANGE))*RAIN_PER_CYCLE/RAIN_NORMAL_FALL,0.250);
      t=(atemp[x][y]-NATURAL_TEMPERATURE_BASE)/((float)NATURAL_TEMPERATURE_RANGE);
      if (t<0)
        t=0;
      t=2*pow(t,0.250);
      u=2*sqrt(fabs(rugged[x][y]))*8;
      
//      printf("%f %f %f %f\n",r,t,u,a);
      
      *((float *)(v+0))=(r+t-u)-0.5+jungle_bonus;
      *((float *)(v+5))=(r+t-u*1.5-a)+0.5+swamp_bonus;
      *((float *)(v+10))=(-r*1.5-u/2)+2.9+desert_bonus;
      *((float *)(v+15))=(-u*2+a/4)+3.25+plains_bonus;
      *((float *)(v+20))=(r-u*2)+2.50+grassland_bonus;
      *((float *)(v+25))=(r/2)-0.5+forest_bonus;
      *((float *)(v+30))=(u+a)-4+hills_bonus;
      *((float *)(v+35))=(-t*2-u/2)+4.5+tundra_bonus;
      *((float *)(v+40))=(-t*4+a/4-u/2)+7+glacier_bonus;
      *((float *)(v+45))=(a*a+u)-4.8+mountain_bonus;

      v[4]=0;
      v[9]=3;
      v[14]=6;
      v[19]=9;
      v[24]=12;
      v[29]=15;
      v[34]=18;
      v[39]=21;
      v[44]=24;
      v[49]=27;
      
//      qsort(v,10,5,greaterthan);
      for (i=1; i<10; i++)
        {
        if (*((float *)(v+0))<*((float *)(v+i*5)))
          {
          *((float *)(v+0))=*((float *)(v+i*5));
          *((char *)(v+4))=*((char *)(v+i*5+4));
//          printf("%d ",*((char *)(v+4)));
          }
        }
//      printf("%d\n",v[4]/3);
      
      colormap[x][y]=v[4];
      switch (v[4]/3)
        {
        case 0 :	//Jungle
          {
          for (x2=x; x2<x+1; x2+=1/vr_ratio)
            {
            for (y2=y; y2<y+1; y2+=1/vr_ratio)
              {
              terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*TREE_HEIGHT;
              }
            }
          break;
          }
        case 1 :	//Swamp
          {
          f=terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))];
          for (x2=x; x2<x+1; x2+=1/vr_ratio)
            {
            for (y2=y; y2<y+1; y2+=1/vr_ratio)
              {
              terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=f_random()*f_random()*f_random()*2*TREE_HEIGHT;
              }
            }
          if (fabs(f-terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))])<TREE_HEIGHT/5)
            colormap[x][y]++;
          if (fabs(f-terrain[((int)(x*vr_ratio))*detail+((int)(y*vr_ratio))])<TREE_HEIGHT/30)
            colormap[x][y]++;
          break;
          }
        case 2 :	//Desert
          {
          for (x2=x; x2<x+1; x2+=1/vr_ratio)
            {
            for (y2=y; y2<y+1; y2+=1/vr_ratio)
              {
              terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.4+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.075+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.075+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.075+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.075+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025;
              }
            }
          if (random()&1)
            colormap[x][y]++;
          if (random()&((2<<((int)((1.0/(r+0.1)))))-1)==((2<<((int)((1.0/(r+0.1)))))-1))
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=CACTUS_HEIGHT/(2-colormap[x][y]%3);
                }
              }
            colormap[x][y]++;
            }
          break;
          }
        case 3 :	//Plains
          {
          for (x2=x; x2<x+1; x2+=1/vr_ratio)
            {
            for (y2=y; y2<y+1; y2+=1/vr_ratio)
              {
              terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.2+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.1+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.1+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05;
              }
            }
          if (random()&1)
            {
            colormap[x][y]++;
            if (random()&63==63)
              {
              for (x2=x; x2<x+1; x2+=1/vr_ratio)
                {
                for (y2=y; y2<y+1; y2+=1/vr_ratio)
                  {
                  terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=BOULDER_HEIGHT;
                  terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                  terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                  terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                  terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                  }
                }
              colormap[x][y]++;
              colormap[x_dec][y]=colormap[x][y];
              colormap[x_inc][y]=colormap[x][y];
              colormap[x][y_dec]=colormap[x][y];
              colormap[x][y_inc]=colormap[x][y];
              }
            }
          break;
          }
        case 4 :	//Grassland
          {
          for (x2=x; x2<x+1; x2+=1/vr_ratio)
            {
            for (y2=y; y2<y+1; y2+=1/vr_ratio)
              {
              terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.2+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.1+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.1+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05;
              }
            }
          if (random()&1)
            {
            colormap[x][y]++;
            if (random()&63==63)
              {
              for (x2=x; x2<x+1; x2+=1/vr_ratio)
                {
                for (y2=y; y2<y+1; y2+=1/vr_ratio)
                  {
                  terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=TREE_HEIGHT;
                  terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=TREE_HEIGHT;
                  terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*TREE_HEIGHT;
                  terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*TREE_HEIGHT;
                  terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]+=uf_random()*TREE_HEIGHT;
                  terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]+=uf_random()*TREE_HEIGHT;
                  }
                }
              colormap[x][y]++;
              colormap[x_dec][y]=colormap[x][y];
              colormap[x_inc][y]=colormap[x][y];
              colormap[x][y_dec]=colormap[x][y];
              colormap[x][y_inc]=colormap[x][y];
              }
            }
          break;
          }
        case 5 :	//Forest
          {
          for (x2=x; x2<x+1; x2+=1/vr_ratio)
            {
            for (y2=y; y2<y+1; y2+=1/vr_ratio)
              {
              terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*TREE_HEIGHT;
              }
            }
          if (a+f_random()*0.1>0.25)
            colormap[x][y]++;
          if (a+f_random()*0.1>0.40)
            colormap[x][y]++;
          break;
          }
        case 6 :	//Hills
          {
          if (random()&7)
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.5+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.05+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.05+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.05+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025;
                }
              }
            }
          if (!(random()&7))
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=f_random()*TREE_HEIGHT;
                }
              }
            colormap[x][y]++;
            }
          break;
          }
        case 7 :	//Tundra
          {
          if (random()&7)
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.5+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.05+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.05+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.05+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025;
                }
              }
            }
          if (!(random()&15))
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=f_random()*ROCK_SHEER_SIZE;
                }
              }
            colormap[x][y]++;
            }
          if (random()&15==15)
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]+=BOULDER_HEIGHT;
                terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]+=uf_random()*BOULDER_HEIGHT;
                }
              }
            colormap[x][y]++;
            colormap[x_dec][y]=colormap[x][y];
            colormap[x_inc][y]=colormap[x][y];
            colormap[x][y_dec]=colormap[x][y];
            colormap[x][y_inc]=colormap[x][y];
            }
          break;
          }
        case 8 :	//Glacier
          {
          if (random()&1)
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.5+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.05+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.05+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.05+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.025+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.025;
                }
              }
            }
          if (rugged[x][y]<0)
            {
            for (x2=x; x2<x+1; x2+=1/vr_ratio)
              {
              for (y2=y; y2<y+1; y2+=1/vr_ratio)
                {
                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.2+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.1+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.1+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05;
                terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]+=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))];
                terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]+=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))];
                terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]+=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))];
                terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]+=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))];
                terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]/=2;
                terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]/=2;
                terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]/=2;
                terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]/=2;
                }
              }
            colormap[x][y]++;
            }
          break;
          }
        case 9 :	//Mountain
          {
          if (fabs(u)+uf_random()*0.9<2.0)
            {
//            for (x2=x; x2<x+1; x2+=1/vr_ratio)
//              {
//              for (y2=y; y2<y+1; y2+=1/vr_ratio)
//                {
//                terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]=terrain[((int)(x2*vr_ratio))*detail+((int)(y2*vr_ratio))]*.2+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y2*vr_ratio))]*.1+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y2*vr_ratio))]*0.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*.1+terrain[((int)(x2*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*.1+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_dec*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_dec*vr_ratio))]*0.05+terrain[((int)(x_inc*vr_ratio))*detail+((int)(y_inc*vr_ratio))]*0.05+f_random()*BOULDER_HEIGHT;
//                }
//              }
            colormap[x][y]++;
            if (t+f_random()*0.05<0.35)
              colormap[x][y]++;
            }
          break;
          }
        }
      }
    }

//  for (x=0; x<REAL_DETAIL; x++)
//    {
//    for (y=0; y<REAL_DETAIL; y++)
//      {
//      printf("%d ",colormap[x][y]);
//      }
//   }
//  printf("\n\n\n");
  return colormap;
  }

char **generate_rivers()
  {
  }

short **add_resources()			//Duh.  
  {
  }

void draw_fractal_terrain(float *terrain, unsigned int edges, vector **windp, vector **ocean, float **otemp, float **atemp, float **rain, char **colormap, int xloc, int yloc, int zoom, int detail, float width, float height, int o_detail, float o_amp, float o_wavelen1, float o_wavelen2, float o_wavelen3, float o_wavelen4, float o_period, float e_detail, float e_amplitude)
  {
  int x,y,i;
  float a;
  int x_plus,y_plus;
  GLfloat az,bz,nx,ny,nz,r;
  float tw=width/((float)detail)*2;
  float th=height/((float)detail)*2;
  float tz,f1,f2,f3,f4;
  int dzoom;
  
  float o_temp[o_detail*2];	//2 rows of ocean    #NOTE - do not confuse with otemp, which is the ocean temperature and is passed.  This is a temporary variable, and was named before the ocean temperature.  Sorry.    -- The management.

  dzoom=1;
  for (i=0; i<zoom; i++)
    dzoom*=2;

  y=0;
  if (((map_mode==4)||(map_mode==5))&&(KEEP_ATEMP_AROUND))
    {
    y=0;
    for (x=0; x<detail; x++)
      {
      if (x<detail-1)
        x_plus=x+1;
      else
        x_plus=0;

      glBegin(GL_TRIANGLE_STRIP);
        az=terrain[detail*x_plus+1]-terrain[detail*x];
        bz=terrain[detail*x_plus]-terrain[detail*x];
        nx=th*bz-az*0;
        ny=az*tw-tw*bz;
        nz=tw*0-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        f1=atemp[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/15-16;
        i=((int)f1)+11;
        if (i>18)
          i=18;
        else if (i<11)
          i=11;
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);    
        glColor3fv(mat_diffuse[i]);

        glVertex3f((x-detail/2)*tw,(-detail/2)*th,terrain[detail*x+0]);
        glVertex3f((x+1-detail/2)*tw,(-detail/2)*th,terrain[detail*x_plus+0]);
        glVertex3f((x+1-detail/2)*tw,(1-detail/2)*th,terrain[detail*x_plus+1]);

        az=terrain[detail*x+1]-terrain[detail*x+0];
        bz=terrain[detail*x_plus+1]-terrain[detail*x+0];
        nx=th*bz-az*th;
        ny=az*tw-0*bz;
        nz=0*th-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        f1=atemp[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/15-16;
        i=((int)f1)+11;
        if (i>18)
          i=18;
        else if (i<11)
          i=11;
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);    
        glColor3fv(mat_diffuse[i]);

        glVertex3f((x-detail/2)*tw,(0+1-detail/2)*th,terrain[detail*x+1]);

        for (y=0; y<detail; y++)
          {
          if (y<detail-1)
            y_plus=y+1;
          else
            y_plus=0;

          az=terrain[detail*x_plus+y_plus]-terrain[detail*x+y];
          bz=terrain[detail*x_plus+y]-terrain[detail*x+y];
          nx=th*bz-az*0;
          ny=az*tw-tw*bz;
          nz=tw*0-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          f1=atemp[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/15-16;
          i=((int)f1)+11;
          if (i>18)
            i=18;
          else if (i<11)
            i=11;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);    
          glColor3fv(mat_diffuse[i]);

          glVertex3f((x+1-detail/2)*tw,(y+1-detail/2)*th,terrain[detail*x_plus+y_plus]);

          az=terrain[detail*x+y_plus]-terrain[detail*x+y];
          bz=terrain[detail*x_plus+y_plus]-terrain[detail*x+y];
          nx=th*bz-az*th;
          ny=az*tw-0*bz;
          nz=0*th-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          f1=atemp[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/15-16;
          i=((int)f1)+11;
          if (i>18)
            i=18;
          else if (i<11)
            i=11;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
          glColor3fv(mat_diffuse[i]);

          glVertex3f((x-detail/2)*tw,(y+1-detail/2)*th,terrain[detail*x+y_plus]);
          }
      glEnd();
      }
    }
  else if (map_mode==6)
    {
    for (x=0; x<detail; x++)
      {
      y=0;
      if (x<detail-1)
        x_plus=x+1;
      else
        x_plus=0;

      glBegin(GL_TRIANGLE_STRIP);
        az=terrain[detail*x_plus+1]-terrain[detail*x];
        bz=terrain[detail*x_plus]-terrain[detail*x];
        nx=th*bz-az*0;
        ny=az*tw-tw*bz;
        nz=tw*0-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        f1=rain[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/40;
        i=((int)f1)+11;
        if (i>18)
          i=18;
        else if (i<11)
          i=11;
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);    
        glColor3fv(mat_diffuse[i]);

        glVertex3f((x-detail/2)*tw,(-detail/2)*th,terrain[detail*x+0]);
        glVertex3f((x+1-detail/2)*tw,(-detail/2)*th,terrain[detail*x_plus+0]);
        glVertex3f((x+1-detail/2)*tw,(1-detail/2)*th,terrain[detail*x_plus+1]);

        az=terrain[detail*x+1]-terrain[detail*x+0];
        bz=terrain[detail*x_plus+1]-terrain[detail*x+0];
        nx=th*bz-az*th;
        ny=az*tw-0*bz;
        nz=0*th-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        f1=rain[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/40;
        i=((int)f1)+11;
        if (i>18)
          i=18;
        else if (i<11)
          i=11;
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);    
        glColor3fv(mat_diffuse[i]);

        glVertex3f((x-detail/2)*tw,(0+1-detail/2)*th,terrain[detail*x+1]);

        for (y=0; y<detail; y++)
          {
          if (y<detail-1)
            y_plus=y+1;
          else
            y_plus=0;

          az=terrain[detail*x_plus+y_plus]-terrain[detail*x+y];
          bz=terrain[detail*x_plus+y]-terrain[detail*x+y];
          nx=th*bz-az*0;
          ny=az*tw-tw*bz;
          nz=tw*0-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

//          f1=rain[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/40;
//          i=((int)f1)+11;
//          if (i>18)
//            i=18;
//          else if (i<11)
//            i=11;
//          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);    
//          glColor3fv(mat_diffuse[i]);

          glVertex3f((x+1-detail/2)*tw,(y+1-detail/2)*th,terrain[detail*x_plus+y_plus]);

          az=terrain[detail*x+y_plus]-terrain[detail*x+y];
          bz=terrain[detail*x_plus+y_plus]-terrain[detail*x+y];
          nx=th*bz-az*th;
          ny=az*tw-0*bz;
          nz=0*th-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          f1=atemp[(int)(x/vr_ratio)][(int)(y/vr_ratio)]/15-16;
          i=((int)f1)+11;
          if (i>18)
            i=18;
          else if (i<11)
            i=11;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
          glColor3fv(mat_diffuse[i]);

          glVertex3f((x-detail/2)*tw,(y+1-detail/2)*th,terrain[detail*x+y_plus]);
          }
      glEnd();
      }
    }
  else
    {
    for (x=0; x<detail; x++)
      {
      y=0;
      if (x<detail-1)
        x_plus=x+1;
      else
        x_plus=0;

      glBegin(GL_TRIANGLE_STRIP);
        az=terrain[detail*x_plus+1]-terrain[detail*x];
        bz=terrain[detail*x_plus]-terrain[detail*x];
        nx=th*bz-az*0;
        ny=az*tw-tw*bz;
        nz=tw*0-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        if (JUNGLE_MOUNTAINS_ONLY)
          {
          if (nz>-0.45)
            {
            glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[46]);
            glColor3fv(mat_diffuse[46]);
            }
          else
            {
            glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[19]);
            glColor3fv(mat_diffuse[19]);
            }
          }
        else
          {
//  printf("1 - (%d %d) (%d %d)  %f %d %d   %d %d\n",x,y,xloc,yloc,vr_ratio,detail,dzoom,(int)((xloc-detail/dzoom/2+x/dzoom)/vr_ratio),(int)((yloc-detail/dzoom/2+y/dzoom)/vr_ratio));
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[19+colormap[(int)((xloc-detail/dzoom/2+x/dzoom)/vr_ratio)][(int)((yloc-detail/dzoom/2+y/dzoom)/vr_ratio)]]);
//  printf("2\n");
          glColor3fv(mat_diffuse[19+colormap[(int)((xloc-detail/dzoom/2+x/dzoom/2)/vr_ratio)][(int)((yloc-detail/dzoom/2+y/dzoom/2)/vr_ratio)]]);
//  printf("3\n");
          }

        glVertex3f((x-detail/2)*tw,(-detail/2)*th,terrain[detail*x+0]);
        glVertex3f((x+1-detail/2)*tw,(-detail/2)*th,terrain[detail*x_plus+0]);
        glVertex3f((x+1-detail/2)*tw,(1-detail/2)*th,terrain[detail*x_plus+1]);

        az=terrain[detail*x+1]-terrain[detail*x+0];
        bz=terrain[detail*x_plus+1]-terrain[detail*x+0];
        nx=th*bz-az*th;
        ny=az*tw-0*bz;
        nz=0*th-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f((x-detail/2)*tw,(0+1-detail/2)*th,terrain[detail*x+1]);

        for (y=0; y<detail; y++)
          {
          if (y<detail-1)
            y_plus=y+1;
          else
            y_plus=0;

          az=terrain[detail*x_plus+y_plus]-terrain[detail*x+y];
          bz=terrain[detail*x_plus+y]-terrain[detail*x+y];
          nx=th*bz-az*0;
          ny=az*tw-tw*bz;
          nz=tw*0-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          if (JUNGLE_MOUNTAINS_ONLY)
            {
            if (nz>-0.45)
              {
              glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[46]);
              glColor3fv(mat_diffuse[46]);
              }
            else
              {
              glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[19]);
              glColor3fv(mat_diffuse[19]);
              }
            }
          else
            {
            glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[19+colormap[(int)((xloc-detail/dzoom/2+x/dzoom)/vr_ratio)][(int)((yloc-detail/dzoom/2+y/dzoom)/vr_ratio)]]);
            glColor3fv(mat_diffuse[19+colormap[(int)((xloc-detail/dzoom/2+x/dzoom)/vr_ratio)][(int)((yloc-detail/dzoom/2+y/dzoom)/vr_ratio)]]);
            }

          glVertex3f((x+1-detail/2)*tw,(y+1-detail/2)*th,terrain[detail*x_plus+y_plus]);

          az=terrain[detail*x+y_plus]-terrain[detail*x+y];
          bz=terrain[detail*x_plus+y_plus]-terrain[detail*x+y];
          nx=th*bz-az*th;
          ny=az*tw-0*bz;
          nz=0*th-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          glVertex3f((x-detail/2)*tw,(y+1-detail/2)*th,terrain[detail*x+y_plus]);
          }
      glEnd();
      }
    }

  //Now, for the ocean.  Generate it at wavelen 0.  It is 2 interfering sin waves.

  tw=width/((float)o_detail)*2;
  th=height/((float)o_detail)*2;
  
  cycle_pos++;
  if (cycle_pos>=o_period)
    cycle_pos-=o_period;

  for (x=0; x<2; x++)
    {
    for (y=0; y<o_detail; y++)
      {
      o_temp[x*o_detail+y]=(sin((o_wavelen1*x+o_wavelen2*y+cycle_pos/o_period)*PI*2)+sin((o_wavelen3*x+o_wavelen4*y+cycle_pos/o_period)*PI*2))*o_amp;
      }
    }

  y=0;
  if (((map_mode==3)||(map_mode==5))&&(KEEP_OTEMP_AROUND))
    {
    f2=((float)detail)/((float)o_detail);
    for (x=0; x<o_detail; x++)
      {
      if (x<o_detail-1)
        x_plus=x+1;
      else
        x_plus=0;

      glBegin(GL_TRIANGLE_STRIP);
        az=o_temp[o_detail*1+1]-o_temp[o_detail*0];
        bz=o_temp[o_detail*1]-o_temp[o_detail*0];
        nx=th*bz-az*0;
        ny=az*tw-tw*bz;
        nz=tw*0-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
      
        f1=otemp[((int)(f2*x/vr_ratio))][((int)(f2*y/vr_ratio))]/15-16;
        i=((int)f1)+11;
        if (i>18)
          i=18;
        else if (i<11)
          i=11;
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
        glColor3fv(mat_diffuse[i]);

        glVertex3f((x-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*0+0]);
        glVertex3f((x+1-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*1+0]);
        glVertex3f((x+1-o_detail/2)*tw,(1-o_detail/2)*th,o_temp[o_detail*1+1]);

        az=o_temp[o_detail*0+1]-o_temp[o_detail*0+0];
        bz=o_temp[o_detail*1+1]-o_temp[o_detail*0+0];
        nx=th*bz-az*th;
        ny=az*tw-0*bz;
        nz=0*th-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f((x-o_detail/2)*tw,(0+1-o_detail/2)*th,o_temp[o_detail*0+1]);

        for (y=0; y<o_detail; y++)
          {
          if (y<o_detail-1)
            y_plus=y+1;
          else
            y_plus=0;

          az=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y]-o_temp[o_detail*0+y];
          nx=th*bz-az*0;
          ny=az*tw-tw*bz;
          nz=tw*0-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          f1=otemp[((int)(f2*x/vr_ratio))][((int)(f2*y/vr_ratio))]/15-16;
          i=((int)f1)+11;
          if (i>18)
            i=18;
          else if (i<11)
            i=11;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
          glColor3fv(mat_diffuse[i]);

          glVertex3f((x+1-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*1+y_plus]);

          az=o_temp[o_detail*0+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          nx=th*bz-az*th;
          ny=az*tw-0*bz;
          nz=0*th-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          glVertex3f((x-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*0+y_plus]);
          }
        for (y=0; y<o_detail; y++)
          {
          o_temp[y]=o_temp[o_detail+y];
          o_temp[o_detail+y]=(sin((o_wavelen1*(x+2)+o_wavelen2*y+cycle_pos/o_period)*PI*2)+sin((o_wavelen3*(x+2)+o_wavelen4*y+cycle_pos/o_period)*PI*2))*o_amp;
          }
      glEnd();
      }
    }
  else if (map_mode==4)
    {
    f2=((float)detail)/((float)o_detail);
    for (x=0; x<o_detail; x++)
      {
      if (x<o_detail-1)
        x_plus=x+1;
      else
        x_plus=0;

      glBegin(GL_TRIANGLE_STRIP);
        az=o_temp[o_detail*1+1]-o_temp[o_detail*0];
        bz=o_temp[o_detail*1]-o_temp[o_detail*0];
        nx=th*bz-az*0;
        ny=az*tw-tw*bz;
        nz=tw*0-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
      
        f1=atemp[((int)(f2*x/vr_ratio))][((int)(f2*y/vr_ratio))]/15-16;
        i=((int)f1)+11;
        if (i>18)
          i=18;
        else if (i<11)
          i=11;
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
        glColor3fv(mat_diffuse[i]);

        glVertex3f((x-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*0+0]);
        glVertex3f((x+1-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*1+0]);
        glVertex3f((x+1-o_detail/2)*tw,(1-o_detail/2)*th,o_temp[o_detail*1+1]);

        az=o_temp[o_detail*0+1]-o_temp[o_detail*0+0];
        bz=o_temp[o_detail*1+1]-o_temp[o_detail*0+0];
        nx=th*bz-az*th;
        ny=az*tw-0*bz;
        nz=0*th-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f((x-o_detail/2)*tw,(0+1-o_detail/2)*th,o_temp[o_detail*0+1]);

        for (y=0; y<o_detail; y++)
          {
          if (y<o_detail-1)
            y_plus=y+1;
          else
            y_plus=0;

          az=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y]-o_temp[o_detail*0+y];
          nx=th*bz-az*0;
          ny=az*tw-tw*bz;
          nz=tw*0-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          f1=atemp[((int)(f2*x/vr_ratio))][((int)(f2*y/vr_ratio))]/15-16;
          i=((int)f1)+11;
          if (i>18)
            i=18;
          else if (i<11)
            i=11;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
          glColor3fv(mat_diffuse[i]);

          glVertex3f((x+1-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*1+y_plus]);

          az=o_temp[o_detail*0+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          nx=th*bz-az*th;
          ny=az*tw-0*bz;
          nz=0*th-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          glVertex3f((x-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*0+y_plus]);
          }
        for (y=0; y<o_detail; y++)
          {
          o_temp[y]=o_temp[o_detail+y];
          o_temp[o_detail+y]=(sin((o_wavelen1*(x+2)+o_wavelen2*y+cycle_pos/o_period)*PI*2)+sin((o_wavelen3*(x+2)+o_wavelen4*y+cycle_pos/o_period)*PI*2))*o_amp;
          }
      glEnd();
      }
    }
  else if (map_mode==6)
    {
    f2=((float)detail)/((float)o_detail);
    for (x=0; x<o_detail; x++)
      {
      if (x<o_detail-1)
        x_plus=x+1;
      else
        x_plus=0;

      glBegin(GL_TRIANGLE_STRIP);
        az=o_temp[o_detail*1+1]-o_temp[o_detail*0];
        bz=o_temp[o_detail*1]-o_temp[o_detail*0];
        nx=th*bz-az*0;
        ny=az*tw-tw*bz;
        nz=tw*0-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
      
        f1=rain[((int)(f2*x/vr_ratio))][((int)(f2*y/vr_ratio))]/40;
        i=((int)f1)+11;
        if (i>18)
          i=18;
        else if (i<11)
          i=11;
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
        glColor3fv(mat_diffuse[i]);

        glVertex3f((x-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*0+0]);
        glVertex3f((x+1-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*1+0]);
        glVertex3f((x+1-o_detail/2)*tw,(1-o_detail/2)*th,o_temp[o_detail*1+1]);

        az=o_temp[o_detail*0+1]-o_temp[o_detail*0+0];
        bz=o_temp[o_detail*1+1]-o_temp[o_detail*0+0];
        nx=th*bz-az*th;
        ny=az*tw-0*bz;
        nz=0*th-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f((x-o_detail/2)*tw,(0+1-o_detail/2)*th,o_temp[o_detail*0+1]);

        for (y=0; y<o_detail; y++)
          {
          if (y<o_detail-1)
            y_plus=y+1;
          else
            y_plus=0;

          az=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y]-o_temp[o_detail*0+y];
          nx=th*bz-az*0;
          ny=az*tw-tw*bz;
          nz=tw*0-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          f1=rain[((int)(f2*x/vr_ratio))][((int)(f2*y/vr_ratio))]/40;
          i=((int)f1)+11;
          if (i>18)
            i=18;
          else if (i<11)
            i=11;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
          glColor3fv(mat_diffuse[i]);

          glVertex3f((x+1-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*1+y_plus]);

          az=o_temp[o_detail*0+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          nx=th*bz-az*th;
          ny=az*tw-0*bz;
          nz=0*th-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          glVertex3f((x-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*0+y_plus]);
          }
        for (y=0; y<o_detail; y++)
          {
          o_temp[y]=o_temp[o_detail+y];
          o_temp[o_detail+y]=(sin((o_wavelen1*(x+2)+o_wavelen2*y+cycle_pos/o_period)*PI*2)+sin((o_wavelen3*(x+2)+o_wavelen4*y+cycle_pos/o_period)*PI*2))*o_amp;
          }
      glEnd();
      }
    }
  else if (map_mode==MAX_MAP_MODE)
    {
    //Don't draw the ocean
    }
  else
    {
    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[2]);
    glColor3fv(mat_diffuse[2]);
    for (x=0; x<o_detail; x++)
      {
      if (x<o_detail-1)
        x_plus=x+1;
      else
        x_plus=0;

      glBegin(GL_TRIANGLE_STRIP);
        az=o_temp[o_detail*1+1]-o_temp[o_detail*0];
        bz=o_temp[o_detail*1]-o_temp[o_detail*0];
        nx=th*bz-az*0;
        ny=az*tw-tw*bz;
        nz=tw*0-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
      

        glVertex3f((x-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*0+0]);
        glVertex3f((x+1-o_detail/2)*tw,(-o_detail/2)*th,o_temp[o_detail*1+0]);
        glVertex3f((x+1-o_detail/2)*tw,(1-o_detail/2)*th,o_temp[o_detail*1+1]);

        az=o_temp[o_detail*0+1]-o_temp[o_detail*0+0];
        bz=o_temp[o_detail*1+1]-o_temp[o_detail*0+0];
        nx=th*bz-az*th;
        ny=az*tw-0*bz;
        nz=0*th-th*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f((x-o_detail/2)*tw,(0+1-o_detail/2)*th,o_temp[o_detail*0+1]);

        for (y=0; y<o_detail; y++)
          {
          if (y<o_detail-1)
            y_plus=y+1;
          else
            y_plus=0;

          az=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y]-o_temp[o_detail*0+y];
          nx=th*bz-az*0;
          ny=az*tw-tw*bz;
          nz=tw*0-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          glVertex3f((x+1-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*1+y_plus]);

          az=o_temp[o_detail*0+y_plus]-o_temp[o_detail*0+y];
          bz=o_temp[o_detail*1+y_plus]-o_temp[o_detail*0+y];
          nx=th*bz-az*th;
          ny=az*tw-0*bz;
          nz=0*th-th*tw;
          r=sqrt(nx*nx+ny*ny+nz*nz);
          if (r>=0.000001) //avoid division by zero
            {
            nx/=r;
            ny/=r;
            nz/=r;
            }
          else
            {
            nx=0;
            ny=0;
            nz=-1;
            }
          glNormal3f(nx,ny,nz);

          glVertex3f((x-o_detail/2)*tw,(y+1-o_detail/2)*th,o_temp[o_detail*0+y_plus]);
          }
//        glVertex3f((x-o_detail/2)*tw,(o_detail/2)*th,10);
//        glVertex3f((x-o_detail/2)*tw,(o_detail/2)*th,10);
//        glVertex3f((x-o_detail/2)*tw,(-o_detail/2)*th,10);
//        glVertex3f((x-o_detail/2)*tw,(-o_detail/2)*th,10);
        for (y=0; y<o_detail; y++)
          {
          o_temp[y]=o_temp[o_detail+y];
          o_temp[o_detail+y]=(sin((o_wavelen1*(x+2)+o_wavelen2*y+cycle_pos/o_period)*PI*2)+sin((o_wavelen3*(x+2)+o_wavelen4*y+cycle_pos/o_period)*PI*2))*o_amp;
          }
      glEnd();
      }
    }

  tw=width/((float)detail)*2;
  th=height/((float)detail)*2;

  //Now for the edges.  Use the random seed, then go around the edges and make them rocky.
  //edges are DISABLED due to bugs.  Sorry.
/*
  srandom(edges);

  tz=sqrt(width*height)/((float)e_detail)*2;
  
  for (x=0; x<detail; x++)
    {
    if (x_plus<detail-1)
      x_plus=x+1;
    else
      x_plus=0;
    if (terrain[x*detail+0]<0)
      {
      glBegin(GL_QUADS);
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[2]);
        glColor3fv(mat_diffuse[2]);
        glNormal3f(0,1,0);
        glVertex3f((x-detail/2)*tw,-detail/2*th,0);
        glVertex3f((x+1-detail/2)*tw,-detail/2*th,0);
        glVertex3f((x+1-detail/2)*tw,-detail/2*th,terrain[x_plus*detail+0]);
        glVertex3f((x-detail/2)*tw,-detail/2*th,terrain[x*detail+0]);
      glEnd();
      }
    if (terrain[x*detail+detail-1]<0)
      {
      glBegin(GL_QUADS);
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[2]);
        glColor3fv(mat_diffuse[2]);
        glNormal3f(0,-1,0);
        glVertex3f((x-detail/2)*tw,detail/2*th+th,0);
        glVertex3f((x+1-detail/2)*tw,detail/2*th+th,0);
        glVertex3f((x+1-detail/2)*tw,detail/2*th+th,terrain[x_plus*detail+detail-1]);
        glVertex3f((x-detail/2)*tw,detail/2*th+th,terrain[x*detail+detail-1]);
      glEnd();
      }
    glBegin(GL_TRIANGLE_STRIP);
      glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[1]);
      glColor3fv(mat_diffuse[1]);

      f1=f_random()*e_amplitude;
      f2=f_random()*e_amplitude;
      glVertex3f((x-detail/2)*tw,f2-detail/2*th,0);
      glVertex3f((x+1-detail/2)*tw,f1-detail/2*th,0);
      for (a=0; ((a>terrain[x*detail+0])&&(a>terrain[x_plus*detail+0])); a-=tz)
        {
        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        
        az=tz;
        bz=0;
        nx=(f2-f1)*bz-az*(f3-f1);
        ny=az*tw-0*bz;
        nz=0*(f3-f1)-(f2-f1)*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
        
        glVertex3f((x-detail/2)*tw,f1-detail/2*th,a);

        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        az=tz;
        bz=tz;
        nx=(f2-f1)*bz-az*(f3-f1);
        ny=az*tw-tw*bz;
        nz=tw*(f3-f1)-(f2-f1)*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f((x+1-detail/2)*tw,f1-detail/2*th,a);
        }
      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      
      az=tz;
      bz=0;
      nx=(f2-f1)*bz-az*(f3-f1);
      ny=az*tw-0*bz;
      nz=0*(f3-f1)-(f2-f1)*tw;
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);
      
      glVertex3f((x-detail/2)*tw,f1-detail/2*th,terrain[x*detail+0]);

      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      az=tz;
      bz=tz;
      nx=(f2-f1)*bz-az*(f3-f1);
      ny=az*tw-tw*bz;
      nz=tw*(f3-f1)-(f2-f1)*tw;
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);

      glVertex3f((x+1-detail/2)*tw,f1-detail/2*th,terrain[(x+1)*detail+0]);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
      glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[1]);
      glColor3fv(mat_diffuse[1]);
      f1=f_random()*e_amplitude;
      f2=f_random()*e_amplitude;
      glVertex3f((x-detail/2)*tw,f2+detail/2*th,0);
      glVertex3f((x+1-detail/2)*tw,f1+detail/2*th,0);
      for (a=0; ((a>terrain[x*detail+detail-1])&&(a>terrain[x_plus*detail+detail-1])); a-=tz)
        {
        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        
        az=tz;
        bz=0;
        nx=(f1-f2)*bz-az*(f1-f3);
        ny=az*tw-0*bz;
        nz=0*(f1-f3)-(f1-f2)*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
        
        glVertex3f((x-detail/2)*tw,f1+detail/2*th,a);

        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        az=tz;
        bz=tz;
        nx=(f1-f2)*bz-az*(f1-f3);
        ny=az*tw-tw*bz;
        nz=tw*(f1-f3)-(f1-f2)*tw;
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f((x+1-detail/2)*tw,f1+detail/2*th,a);
        }
      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      
      az=tz;
      bz=0;
      nx=(f1-f2)*bz-az*(f1-f3);
      ny=az*tw-0*bz;
      nz=0*(f1-f3)-(f1-f2)*tw;
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);
        
      glVertex3f((x-detail/2)*tw,f1+detail/2*th,terrain[x*detail+detail-1]);

      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      az=tz;
      bz=tz;
      nx=(f1-f2)*bz-az*(f1-f3);
      ny=az*tw-tw*bz;
      nz=tw*(f1-f3)-(f1-f2)*tw;
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);

      glVertex3f((x+1-detail/2)*tw,f1+detail/2*th,terrain[(x+1)*detail+detail-1]);
    glEnd();
    }
  
  for (y=0; y<detail; y++)
    {
    if (y<detail-1)
      y_plus=y+1;
    else
      y_plus=0;
    if (terrain[0*detail+y]<0)
      {
      glBegin(GL_QUADS);
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[2]);
        glColor3fv(mat_diffuse[2]);
        glNormal3f(1,0,0);
        glVertex3f(-detail/2*tw,(y-detail/2)*th,0);
        glVertex3f(-detail/2*tw,(y+1-detail/2)*th,0);
        glVertex3f(-detail/2*tw,(y+1-detail/2)*th,terrain[x_plus*detail+0]);
        glVertex3f(-detail/2*tw,(y-detail/2)*th,terrain[x*detail+0]);
      glEnd();
      }
    if (terrain[(detail-1)*detail+y]<0)
      {
      glBegin(GL_QUADS);
        glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[2]);
        glColor3fv(mat_diffuse[2]);
        glNormal3f(-1,0,0);
        glVertex3f((x-detail/2)*tw,(y+detail/2)*th+th,0);
        glVertex3f((x+1-detail/2)*tw,(y+1+detail/2)*th+th,0);
        glVertex3f((x+1-detail/2)*tw,(y+1+detail/2)*th+th,terrain[x_plus*detail+detail-1]);
        glVertex3f((x-detail/2)*tw,(y+detail/2)*th+th,terrain[x*detail+detail-1]);
      glEnd();
      }
    glBegin(GL_TRIANGLE_STRIP);
      glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[1]);
      glColor3fv(mat_diffuse[1]);
      f1=f_random()*e_amplitude;
      f2=f_random()*e_amplitude;
      glVertex3f(f2-detail/2*tw,(y-detail/2)*th,0);
      glVertex3f(f1-detail/2*tw,(y+1-detail/2)*th,0);
      for (a=0; ((a<terrain[0*detail+y])&&(a<terrain[0*detail+y_plus])); a+=tz)
        {
        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        
        az=tz;
        bz=0;
        nx=th*bz-az*0;
        ny=az*(f2-f1)-(f3-f1)*bz;
        nz=(f3-f1)*0-th*(f2-f1);
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
        
        glVertex3f(f1-detail/2*tw,(y-detail/2)*th,a);

        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        
        az=tz;
        bz=tz;
        nx=th*bz-az*th;
        ny=az*(f2-f1)-(f3-f1)*bz;
        nz=(f3-f1)*th-th*(f2-f1);
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f(f1-detail/2*tw,(y+1-detail/2)*th,a);
        }
      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      
      az=tz;
      bz=0;
      nx=th*bz-az*0;
      ny=az*(f2-f1)-(f3-f1)*bz;
      nz=(f3-f1)*0-th*(f2-f1);
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);
        
      glVertex3f(f1-detail/2*tw,(y-detail/2)*th,terrain[x*detail+0]);

      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      
      az=tz;
      bz=tz;
      nx=th*bz-az*th;
      ny=az*(f2-f1)-(f3-f1)*bz;
      nz=(f3-f1)*th-th*(f2-f1);
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);

      glVertex3f(f1-detail/2*tw,(y+1-detail/2)*th,terrain[(x+1)*detail+0]);
    glEnd();
    glBegin(GL_TRIANGLE_STRIP);
      glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[1]);
      glColor3fv(mat_diffuse[1]);
      f1=f_random()*e_amplitude;
      f2=f_random()*e_amplitude;
      glVertex3f(f2+detail/2*tw,(y-detail/2)*th,0);
      glVertex3f(f1+detail/2*tw,(y+1-detail/2)*th,0);
      for (a=0; ((a<terrain[(detail-1)*detail+y])&&(a<terrain[(detail-1)*detail+y_plus])); a+=tz)
        {
        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        
        az=tz;
        bz=0;
        nx=th*bz-az*0;
        ny=az*(f1-f2)-(f1-f3)*bz;
        nz=(f1-f3)*0-th*(f1-f2);
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);
        
        glVertex3f(f1+detail/2*tw,(y-detail/2)*th,a);

        f3=f2;
        f2=f1;
        f1=f_random()*e_amplitude;
        
        az=tz;
        bz=tz;
        nx=th*bz-az*th;
        ny=az*(f1-f2)-(f1-f3)*bz;
        nz=(f1-f3)*th-th*(f1-f2);
        r=sqrt(nx*nx+ny*ny+nz*nz);
        if (r>=0.000001) //avoid division by zero
          {
          nx/=r;
          ny/=r;
          nz/=r;
          }
        else
          {
          nx=0;
          ny=0;
          nz=-1;
          }
        glNormal3f(nx,ny,nz);

        glVertex3f(f1+detail/2*tw,(y+1-detail/2)*th,a);
        }
      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      
      az=tz;
      bz=0;
      nx=th*bz-az*0;
      ny=az*(f1-f2)-(f1-f3)*bz;
      nz=(f1-f3)*0-th*(f1-f2);
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);
      
      glVertex3f(f1+detail/2*tw,(y-detail/2)*th,terrain[x*detail+0]);

      f3=f2;
      f2=f1;
      f1=f_random()*e_amplitude;
      
      az=tz;
      bz=tz;
      nx=th*bz-az*th;
      ny=az*(f1-f2)-(f1-f3)*bz;
      nz=(f1-f3)*th-th*(f1-f2);
      r=sqrt(nx*nx+ny*ny+nz*nz);
      if (r>=0.000001) //avoid division by zero
        {
        nx/=r;
        ny/=r;
        nz/=r;
        }
      else
        {
        nx=0;
        ny=0;
        nz=-1;
        }
      glNormal3f(nx,ny,nz);

      glVertex3f(f1+detail/2*tw,(y+1-detail/2)*th,terrain[(x+1)*detail+0]);
    glEnd();
    }
  
  srandom(time(0));
*/
  if ((map_mode==1)&&(KEEP_WIND_AROUND))
    {
    //Draw wind arrows.
    
    f2=sqrt(width*height);
    for (x=0; x<detail; x+=ARROW_DRAW_INTERVAL)
      {
      for (y=0; y<detail; y+=ARROW_DRAW_INTERVAL)
        {
        glBegin(GL_TRIANGLES);
          f1=windp[(int)(x/vr_ratio)][(int)(y/vr_ratio)].speed*300/f2;
          i=((int)f1)+3;
          if (i>10)
            i=10;	//Have it put the arrows in the right place, silly.
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
          glColor3fv(mat_diffuse[i]);
          glNormal3f(0,0,-1);
          f3=(x-detail/2+sin(windp[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle)*ARROW_DRAW_INTERVAL*0.5)*tw;
          f4=(y-detail/2+cos(windp[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle)*ARROW_DRAW_INTERVAL*0.5)*th;
          glVertex3f(f3,f4,f2/7);
          f3=(x-detail/2+sin(windp[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*11.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*tw;
          f4=(y-detail/2+cos(windp[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*11.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*th;
          glVertex3f(f3,f4,f2/7);
          f3=(x-detail/2+sin(windp[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*9.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*tw;
          f4=(y-detail/2+cos(windp[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*9.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*th;
          glVertex3f(f3,f4,f2/7);
        glEnd();
        }
      }
    }
  else if ((map_mode==2)&&(KEEP_OCEAN_AROUND))
    {
    //Draw ocean arrows.
    
    f2=sqrt(width*height);
    for (x=0; x<detail; x+=ARROW_DRAW_INTERVAL)
      {
      for (y=0; y<detail; y+=ARROW_DRAW_INTERVAL)
        {
        if (terrain[x*detail+y]<0)
          {
          glBegin(GL_TRIANGLES);
            f1=ocean[(int)(x/vr_ratio)][(int)(y/vr_ratio)].speed*2000/f2;
            i=((int)f1)+3;
            if (i>10)
              i=10;	//Have it put the arrows in the right place, silly.
            glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient[i]);
            glColor3fv(mat_diffuse[i]);
            glNormal3f(0,0,-1);
            f3=(x-detail/2+sin(ocean[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle)*ARROW_DRAW_INTERVAL*0.5)*tw;
            f4=(y-detail/2+cos(ocean[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle)*ARROW_DRAW_INTERVAL*0.5)*th;
            glVertex3f(f3,f4,o_amp*10);
            f3=(x-detail/2+sin(ocean[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*11.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*tw;
            f4=(y-detail/2+cos(ocean[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*11.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*th;
            glVertex3f(f3,f4,o_amp*10);
            f3=(x-detail/2+sin(ocean[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*9.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*tw;
            f4=(y-detail/2+cos(ocean[(int)(x/vr_ratio)][(int)(y/vr_ratio)].angle+(PI*9.0)/10.0)*ARROW_DRAW_INTERVAL*0.5)*th;
            glVertex3f(f3,f4,o_amp*10);
          glEnd();
          }
        }
      }
    }
  return;
  }

float dist_2d(float x1, float y1, float x2, float y2)
  {
  return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
  }

float dist_3d(float x1, float y1, float z1, float x2, float y2, float z2)
  {
  return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
  }

float altitude_at(float *terrain, float width, float height, int detail, float x, float y)
  {
  float nx,ny;
  float d1,d2,d3,dsum;
  float w1,w2,w3,w4;
  float tw,th;
  tw=width/detail;
  th=height/detail;
  nx=x/tw;
  ny=y/th;
  if (ny-(float)((int)ny)>nx-(float)((int)nx))
    {
    d1=dist_2d(nx,ny,(int)nx,(int)ny);
    d2=dist_2d(nx,ny,(int)nx,(int)ny+1);
    d3=dist_2d(nx,ny,(int)nx+1,(int)ny+1);
    dsum=d1+d2+d3;
    return (terrain[detail*((int)nx)+(int)ny]*d1+terrain[detail*((int)nx)+(int)ny+1]*d2+terrain[detail*((int)nx+1)+(int)ny+1]*d3)/dsum;
    }
  else
    {
    d1=dist_2d(nx,ny,(int)nx,(int)ny);
    d2=dist_2d(nx,ny,(int)nx,(int)ny+1);
    d3=dist_2d(nx,ny,(int)nx+1,(int)ny+1);
    dsum=d1+d2+d3;
    return -(terrain[detail*((int)nx)+(int)ny]*d1+terrain[detail*((int)nx)+(int)ny+1]*d2+terrain[detail*((int)nx+1)+(int)ny+1]*d3)/dsum;
    }
  }






