// D E F I N E S //////////////////////////////////////////////////////////////

#define PI				3.1415926535	//Exact value of Pi

#define REAL_DETAIL			256
#define KEEP_WIND_AROUND		1
#define KEEP_OCEAN_AROUND		1
#define KEEP_OTEMP_AROUND		1
#define KEEP_ATEMP_AROUND		1
#define KEEP_RAIN_AROUND		1
#define ARROW_DRAW_INTERVAL		12
#define MAX_MAP_MODE			7
#define MOUNTAIN_SCALAR			500
#define MAX_ZOOM			3

#define JUNGLE_MOUNTAINS_ONLY		0

#define OCEAN_ANGLE_COUNT		9
#define OCEAN_SPEED_COUNT		12
#define OCEAN_SPEED_MAXBR		50
#define OCEAN_SPEED_DIVISOR		8000
#define WESTWARD_INTENSIFICATION	1.0	//if you don't know, look it up
#define NATURAL_TEMPERATURE_RANGE	120	//120 kelvins range
#define NATURAL_TEMPERATURE_BASE	215	//with a min. of 215 kelvins (making it go from 215 to 335 kelvins)
#define OCEAN_MIXING_PASSES		18
#define O_HEAT_DISTRIBUTION_CONSTANT	2.5
#define ATM_MIXING_PASSES		18
#define A_HEAT_DISTRIBUTION_CONSTANT	5.0
#define OCEAN_HEAT_TRANSFER		0.5
#define HEIGHT_HEAT_CONST		600.0
#define DEWPOINT_CONSTANT		0.8	//Used to determine how much moisture the air can hold, relative to its temperature, in centimeters of potential rainfall.
#define RAIN_MIXING_PASSES		18	//Assume each cycle is a year.
#define RAIN_DISTRIBUTION_CONSTANT	1.0	//Rain moves quickly every pass
#define RAIN_PER_CYCLE			110.0
#define RAIN_NORMAL_FALL		1500.0	//How many centimeters of rain in the air when rainfall meets the previously specified amount
#define RAIN_MIN_PERCENT		0.1	//What is the lowest fraction of atmospheric moisture that can rain per cycle
#define ROUGHNESS_SIZE			1

#define TREE_HEIGHT			1.0
#define CACTUS_HEIGHT			1.0
#define BOULDER_HEIGHT			0.5
#define ROCK_SHEER_SIZE			3.0

#define FUTILE				-1

// G L O B A L S //////////////////////////////////////////////////////////////

GLfloat light_pos[]=    { 530.7, -993.1, -2002.3 };
GLfloat light_ambient[]={ 0.10, 0.10, 0.10, 1.00 };
GLfloat light_diffuse[]={ 1.90, 1.90, 1.90, 1.00 };
GLfloat light_specular[]={0.00, 0.00, 0.00, 0.00 };
GLfloat mat_ambient[][4]={	{ 0.080, 0.240, 0.080, 1.00 },	//Unused
				{ 0.400, 0.440, 0.320, 1.00 },	//Unused
				{ 0.080, 0.280, 0.280, 1.00 },	//Ocean
				{ 0.100, 0.100, 0.100, 1.00 },	//Arrow colors
				{ 0.200, 0.200, 0.200, 1.00 },
				{ 0.300, 0.300, 0.300, 1.00 },
				{ 0.400, 0.400, 0.400, 1.00 },
				{ 0.500, 0.500, 0.500, 1.00 },
				{ 0.600, 0.600, 0.600, 1.00 },
				{ 0.700, 0.700, 0.700, 1.00 },
				{ 0.800, 0.800, 0.800, 1.00 },
				{ 0.000, 0.000, 0.000, 1.00 },	//Heat/rain colors
				{ 0.300, 0.000, 0.500, 1.00 },
				{ 0.000, 0.000, 0.800, 1.00 },
				{ 0.000, 0.800, 0.000, 1.00 },
				{ 0.400, 0.400, 0.000, 1.00 },
				{ 0.600, 0.200, 0.000, 1.00 },
				{ 0.800, 0.000, 0.000, 1.00 },
				{ 0.800, 0.400, 0.400, 1.00 },
				{ 0.080, 0.240, 0.080, 1.00 },	//Jungle
				{ 0.080, 0.240, 0.080, 1.00 },
				{ 0.080, 0.240, 0.080, 1.00 },
				{ 0.080, 0.180, 0.120, 1.00 },	//Swamp
				{ 0.080, 0.160, 0.160, 1.00 },
				{ 0.200, 0.200, 0.080, 1.00 },
				{ 0.350, 0.350, 0.200, 1.00 },	//Desert
				{ 0.400, 0.350, 0.200, 1.00 },
				{ 0.120, 0.300, 0.080, 1.00 },
				{ 0.270, 0.230, 0.080, 1.00 },	//Plains
				{ 0.230, 0.270, 0.080, 1.00 },
				{ 0.350, 0.400, 0.300, 1.00 },
				{ 0.100, 0.300, 0.100, 1.00 },	//Grassland
				{ 0.150, 0.320, 0.100, 1.00 },
				{ 0.080, 0.260, 0.080, 1.00 },
				{ 0.150, 0.300, 0.080, 1.00 },	//Forest
				{ 0.140, 0.320, 0.150, 1.00 },
				{ 0.120, 0.280, 0.200, 1.00 },
				{ 0.120, 0.250, 0.120, 1.00 },	//Hills
				{ 0.300, 0.350, 0.250, 1.00 },
				{ 0.350, 0.400, 0.300, 1.00 },
				{ 0.250, 0.200, 0.080, 1.00 },	//Tundra
				{ 0.450, 0.450, 0.500, 1.00 },
				{ 0.350, 0.400, 0.300, 1.00 },
				{ 0.450, 0.450, 0.500, 1.00 },	//Glacier
				{ 0.350, 0.400, 0.300, 1.00 },
				{ 0.350, 0.400, 0.300, 1.00 },
				{ 0.350, 0.400, 0.300, 1.00 },	//Mountain
				{ 0.080, 0.250, 0.080, 1.00 },
				{ 0.450, 0.450, 0.500, 1.00 }	};

GLfloat mat_diffuse[][4]={	{ 0.100, 0.250, 0.100, 1.00 },	//Unused
				{ 0.400, 0.350, 0.300, 1.00 },	//Unused
				{ 0.180, 0.320, 0.350, 1.00 },	//Ocean
				{ 0.100, 0.100, 0.100, 1.00 },	//Arrow colors
				{ 0.200, 0.200, 0.200, 1.00 },
				{ 0.300, 0.300, 0.300, 1.00 },
				{ 0.400, 0.400, 0.400, 1.00 },
				{ 0.500, 0.500, 0.500, 1.00 },
				{ 0.600, 0.600, 0.600, 1.00 },
				{ 0.700, 0.700, 0.700, 1.00 },
				{ 0.800, 0.800, 0.800, 1.00 },
				{ 0.000, 0.000, 0.000, 1.00 },	//Heat/Rain colors
				{ 0.300, 0.000, 0.500, 1.00 },
				{ 0.000, 0.000, 0.800, 1.00 },
				{ 0.000, 0.800, 0.000, 1.00 },
				{ 0.400, 0.400, 0.000, 1.00 },
				{ 0.600, 0.200, 0.000, 1.00 },
				{ 0.800, 0.000, 0.000, 1.00 },
				{ 0.800, 0.400, 0.400, 1.00 },
				{ 0.100, 0.250, 0.100, 1.00 },	//Jungle
				{ 0.100, 0.250, 0.100, 1.00 },
				{ 0.100, 0.250, 0.100, 1.00 },
				{ 0.100, 0.210, 0.150, 1.00 },	//Swamp
				{ 0.100, 0.190, 0.190, 1.00 },
				{ 0.240, 0.240, 0.100, 1.00 },
				{ 0.420, 0.400, 0.240, 1.00 },	//Desert
				{ 0.460, 0.380, 0.240, 1.00 },
				{ 0.160, 0.350, 0.100, 1.00 },
				{ 0.160, 0.350, 0.100, 1.00 },	//Plains
				{ 0.310, 0.270, 0.100, 1.00 },
				{ 0.270, 0.310, 0.100, 1.00 },
				{ 0.140, 0.340, 0.140, 1.00 },	//Grassland
				{ 0.220, 0.370, 0.120, 1.00 },
				{ 0.100, 0.300, 0.100, 1.00 },
				{ 0.180, 0.360, 0.100, 1.00 },	//Forest
				{ 0.170, 0.370, 0.180, 1.00 },
				{ 0.140, 0.320, 0.230, 1.00 },
				{ 0.150, 0.300, 0.150, 1.00 },	//Hills
				{ 0.350, 0.300, 0.250, 1.00 },
				{ 0.400, 0.350, 0.300, 1.00 },
				{ 0.300, 0.250, 0.100, 1.00 },  //Tundra
				{ 0.800, 0.800, 0.870, 1.00 },
				{ 0.400, 0.350, 0.300, 1.00 },
				{ 0.400, 0.350, 0.300, 1.00 },	//Glacier
				{ 0.800, 0.800, 0.870, 1.00 },
				{ 0.400, 0.350, 0.300, 1.00 },
				{ 0.400, 0.350, 0.300, 1.00 },	//Mountain
				{ 0.100, 0.300, 0.100, 1.00 },
				{ 0.800, 0.800, 0.870, 1.00 }	};


int cycle_pos=0;
int map_mode=0;
float vr_ratio;		//Ratio between the visual and real details.

// P R O T O T Y P E S ////////////////////////////////////////////////////////

void _2dfree(void **var, int width);
float f_random();
float uf_random();
float sf_random();
float suf_random();
float square(float f);
float de_mod(int a, int e, float c1, float c2);
float gen_rand(float ratio, float multiplier);
float alt_mod(float alt, float mod);
int generate_fractal_terrain(float **t1, unsigned int *e1, vector ***w1, vector ***o1, float ***tp1, float ***tp2, float ***r1, char ***c1, int steps, float ratio, float ruggedness, int width, int height, int detail, int x_squares, int y_squares, float constant1, float constant2, float alti_mod, float mount_scale, float min_wind, float wind_scale, float ocean_scale, int av_size, float av_pow1, float av_pow2, float sun_scale, float rain_scale, float jungle_bonus, float swamp_bonus, float desert_bonus, float plains_bonus, float grassland_bonus, float forest_bonus, float hills_bonus, float tundra_bonus, float glacier_bonus, float mountain_bonus);
float *zoom(float *terrain, char **colormap, int detail, int steps, float ruggedness, float alti_mod, int seed, int amount, int xloc, int yloc);
vector find_downslope(float *terrain, int x, int y, int detail, float scale);
vector ang_average(float ang1, float speed1, float ang2, float speed2);
float ang_diff(float ang1, float ang2);
vector **generate_wind_map(float *t1, int detail, float min_wind, float wind_scale, float cnst);
static float current_angle(int x, int y, float *terrain, int detail);
static float current_speed(int x, int y, float *terrain, vector **ocean, int detail);
vector **generate_ocean_map(float *t1, int detail, float ocean_scale, float cnst);
float **generate_ocean_temp(vector **ocean, int detail, float scale);
float **generate_atm_temp(float **otemp, vector **wind, float *terrain, int detail, float scale);
float **generate_rain_map(float **atemp, vector **wind, float *terrain, int detail, float scale);
int greaterthan(const void *a, const void *b);
float **generate_rugged_map(float *terrain, float **rugged, int detail);
char **tweak_terrain(float *terrain, float **atemp, float **rain, int detail, float jungle_bonus, float swamp_bonus, float desert_bonus, float plains_bonus, float grassland_bonus, float forest_bonus, float hills_bonus, float tundra_bonus, float glacier_bonus, float mountain_bonus);
char **generate_rivers();
short **add_resources();
void draw_fractal_terrain(float *terrain, unsigned int edges, vector **windp, vector **ocean, float **otemp, float **atemp, float **rain, char **colormap, int xloc, int yloc, int zoom, int detail, float width, float height, int o_detail, float o_amp, float o_wavelen1, float o_wavelen2, float o_wavelen3, float o_wavelen4, float o_period, float e_detail, float e_amplitude);
float dist_2d(float x1, float y1, float x2, float y2);
float dist_3d(float x1, float y1, float z1, float x2, float y2, float z2);
float altitude_at(float *terrain, float width, float height, int detail, float x, float y);
